// DVdec
// By: Andy Kromkamp kromkamp@my-deja.com
//
// THIS SOURCE CODE IS GIVEN AWAY FREELY, WITH NO GUARANTEES EXPLICIT OR IMPLIED.

/////////////////////////////////////////////////////////////////////////////
//  DVSPOOF(ECC_hamming) v0.1 09/29/2001 liaor@iname.com
// 
//    This version of DVSPOOF has been modified to support forward-ECC.
//  It is NOT compatible with the original DVSPOOF.  This version of DVDEC 
//  only accepts files written by its companion DVENC(ECC).
//
//  Ecc-decoder
//  -----------
//    The ECC-decoder pushes received data onto a 'queue', waits until the
//  queue has filled (72 total frames), requests an ecc_correct cycle, 
//  then passes the corrected-data to output.  The decoding sequence is
//  somewhat tangled, and not quite straightforward.
//
//   1) ReadDVFrame() -> SplitDVFrame() -> ParseStatus() -> add2eccqueue() -> 
//   2) dump_eccblock() -> 
//
//     ReadDVFrame() and SplitDVFrame() have not changed.  The ECC-decoder
//   relies on these two functions to identify duplicate and dropped frames.
//   If the current frame is 'expected', at this point it would originally
//   have been dumped to output file.  Instead, the expected-frame is
//   routed to the ecc-decoder (add2eccqueue).
//     When the queue reaches FULL, the pending ecc_block is ready for the
//   correction-process.  dump_eccblock() writes the entire block of 
//   corrected-data (64 data frames) to the output file.
//
//  Terms
//  -----
//    AVIfile - Output unit for dvspoof.exe
//    AVIframe - one frame (audio + video) 
//        Data is stored only in the video-portion of the AVIframe.
//        The audio-portion is unused.
//    ECCblock - Output unit for the ECC_encoder.  
//        1 eccblock = 72 AVIframes (64 data + 8 parity.)
//        *(the first ECCblock in the AVIfile is 'stretched', due to
//          repitition of AVIframe#0.)
//
//  Source Code changes
//  -------------------
//    The addition of ECC-encoder requires changes to dvspoof.  Here
//  is a summary of changes.
//
//    mkdir_recursive() - not related to the ECC-package at all.
//      This function takes an absolute path, breaks it into its
//      individual path components, and calls 'mkdir' on each component.
//      I added this to fix some problems with the creation of 
//      absolute-path directories.  I encountered these problems
//      with MSVC6 and Win98se.  I did not test on any other platform(s)
//    add2eccqueue() - pushes "data+framestatus" into the ECC_decoder's 
//       buffer.  (The ECC_decoder doesn't know about framestatus, so
//       add2eccqueue manually tracks these headers.  dump_eccblock needs
//       the framestatus headers.)
//       When the dcoder's buffer reaches full, automatically performs
//       the ecc_correction operation and calls dump_eccblock() to
//       write the corrected data to output file.  
//    dump_eccblock() - writes 1 ECC-block's worth of processed data to 
//       the output file.  
//    main() - 
//      the main processing-loop has changed quite a bit.  
//      The main() does a preliminary analysis of AVIframes: Duplicate
//      frames are killed.  Dropped-frames are counted, and the same
//      # 'dummy frames' are pushed into the ECC_decoder (to maintain
//      block-alignment.)  No output data is directly written in this
//      function (see add2eccqueue and dump_eccblock.)
//      Also, some code bugfixes (a few local vars weren't initialized)
//     
//   *** compiled with Microsoft Visual C++ 6.0 pro (with ServicePack5)
//      Microsoft doesn't recognize 'long long', so I have replace these
//      with __int64.  I hope this doesn't cause a problem.
//      Also, the 'sleep()' function doesn't seem to work in MSVC.
//      Accordingly, -throttle support has been removed.
//
//      The program spams the screen with a bunch of debug info.
/////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <time.h>

#include "ecc.h"

char AVIDVframeheader[] = { 
    0x30, 0x30, 0x64, 0x63, 0xc0, 0xd4, 0x01, 0x00, // Represents ASCII "00dc" (AVI video frame tag) + size of DV frame (0001d4c0 = 120,000 bytes)
    0x1F, 0x07, 0x00, 0x3F, 0x68, 0x78, 0x78, 0x78 // start of DV frame
    };

//char DVSpoofKey[] = { 'S', 'P', 'O', 'F' };  // original DVSPOOF
char DVSpoofKey[] = { 'S', 'P', 'O', 'E' };  // DVSPOOF(ECC-Hamming)
    
int INFILECOUNT;
char INFILENAME[255];

#define DROPPED_FRAMES 0x00000000 // + # of dropped frames
#define EXPECTED_FRAME 0x80000000
#define DUPLICATE_FRAME 0x80000001
#define REDUNDANT_FRAME0 0x80000002
#define LOCK_FRAME0 0x80000003
#define RESET_FRAME0 0x80000004

//typedef unsigned long long _bigint; // (MS Visual C++ doesn't like)
typedef __int64 _bigint;

typedef struct  {
  int RealFrame;   // Used for Total Dropped bytes, etc
  unsigned int Checksum;    // Checksum calculated for this frame
  unsigned int StoredChecksum; //Checksum stored in DV file
  _bigint FileSize;    // Bytes Left for this input file
  unsigned int FileCount;   // File Number
  unsigned int TotalFiles;  // Total number of files
  int FrameNum;    // Which Frame #
  unsigned int TotalFrames; // Total Frames for this input file

  int Frame0Count;      // Number of Frame0's detected (3 = locked in)
  int ZeroBytes;   // Number of bytes dropped
  int RealBytes;    // Number of bytes decoded
  unsigned int Status;      // Overall status for this frame, see DEFINES
} FrameInfoType;


//ecc_enc_state g_eccEncoderState = {0};  // force all member-vars to 0
ecc_dec_state g_eccDecoderState = {0};

FrameInfoType *g_ecc_fibuffer; // buffer for headers
int g_ecc_n; // frame# index counter
int g_eccEnable; // flag, 1=enable error-correction, 0=no error-correction
int g_recurse; // flag, 1 = recurse subdirectories

char g_startfilename[255]; // argument for '-startfile'
int g_startfilefound;
FILE * g_logfile = NULL;
FILE * g_outfile = NULL;

//////////////////////////////////////////////////////////////////////////////////////////////////   
// BufferCompare will compare 2 buffers. buffer1 can be larger than buffer2, so offset is used
// to determine where to start looking at buffer1 (ie. pattern-match buffer2 inside buffer1)
// This pattern match is maskable by the mask array (0x00 = dont care, 0x01 = care)
//////////////////////////////////////////////////////////////////////////////////////////////////
int BufferCompare (char * buffer1, char * buffer2, char * mask, int size, int offset) {

  int i;
  int match = 1;

  for (i=0; i<size; i++) { 
    if ( (buffer1[offset+i] != buffer2[i]) && (mask[i] == 0x01) ) {
        match = 0;
    }
  }
  return match;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// ReadDVFrame will seek through the .avi files to find the next DV frame, and then load it in.
// As this function is repeatedly called,it keeps reading, file after file, until it finds a frame 
// or it runs out of files
/////////////////////////////////////////////////////////////////////////////////////////////////////////
int ReadDVFrame(unsigned char * buffer, FILE ** infile, unsigned char * infilename,int inplace, FILE ** g_logfile) {
  
  int i,j;  
  int match = 0;
  char headermask[] = {
    0x01, 0x01, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 
    0x00, 0x01, 0x01, 0x01, 0x00, 0x01, 0x01, 0x01
    };
  int tempincr,carry;

  //printf ("\nReadDVFrame() : starting g_ecc_n = %d", g_ecc_n );

  while (1) {
    if (*infile == NULL) {
      *infile = fopen (infilename, "rb"); 
      if (*infile == NULL) { 
        return (0); 
      }
      printf ("\n\n\nReadDVFrame() : (Reading from %s) ",infilename);
      fprintf (*g_logfile, "ReadDVFrame() : \n\n\n(Reading from %s) ",infilename);
      INFILECOUNT++;
    }
    if (fread(buffer,1,120000,*infile) == 120000) {
      //printf("\nReadDVFrame(): got 120,000bytes, g_ecc_n=%d...", g_ecc_n );
      for (i=0; i<119984; i++) {
        if (BufferCompare(buffer,AVIDVframeheader, headermask, 16, i)) { // Found a match, read in whole 120000 bytes DV frame
          fseek(*infile,-119992+i,SEEK_CUR);
          fread(buffer,120000,1,*infile);
          return 1;
        
        }
      }
      // Did not find any frame in this 120kbytes
      fseek(*infile,-50,SEEK_CUR); // Rewind a bit to read another buffer worth of data
    }
    else { // There is less than 120kbytes in this file, so there are no more frames, close file and update filename
      printf("\nReadDVFrame(): end...(g_ecc_n=%d)", g_ecc_n );
      fclose (*infile);
      if (inplace) { // Inplace deletes .avi files as they are finished being used
        remove (infilename);
      }
    
      // Update file name -> pick # out of current filename, increment it and put it back
      carry=1;
      for (j=strlen(infilename)-1;j>=0;j--) {

        if ((infilename[j] > 47) && (infilename[j] < 58)) { // Its a numeric
          tempincr = infilename[j] - 48; // convert to integer
          if (carry == 1) {
            tempincr++;
            carry = 0;
          }
          if (tempincr == 10) {
            tempincr = 0;
            carry = 1;
          }
          infilename[j] = tempincr + 48; // put it back into the string
        }
      }
      *infile = NULL; 
    }
  }
  //printf ("\nReadDVFrame() : exiting");
}




///////////////////////////////////////////////////////////////////////////////////////////////////////////
// SplitDVFrame will filter out the DV structures, and decode the first Video DIF blocks (DVSpoof headers)
///////////////////////////////////////////////////////////////////////////////////////////////////////////
int SplitDVFrame(unsigned char * inbuffer, unsigned char * outbuffer, FrameInfoType * status) {

  int i,j,k;
  unsigned char DVSpoofBuffer1[24], DVSpoofBuffer2[24], DVSpoofBuffer3[24];
  int inbufferoffset;
  int outbufferoffset=0;
  unsigned int checksumtemp=0;

  char mask[] = {
    0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
    0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
    0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01
    };


  // Parse DV Frame, discard DV Structures, Copy DVSpoof Header to a separate buffer and
  // Copy the rest of the real data into a continuous output buffer
  for (i=0; i<10; i++) { // 10 DIF sequences/NTSC frame, each is 12kbytes
    for (j=0; j<150; j++) { // 150 DIF blocks/DIF sequence

      inbufferoffset = (i*12000)+(j*80);
      if ( (j == 7) && (i == 0) ) { // This is the DVSpoof Header block
        for (k=4; k<28;k++) { DVSpoofBuffer1[k-4] = inbuffer[inbufferoffset+k];}
        for (k=28; k<52;k++) { DVSpoofBuffer2[k-28] = inbuffer[inbufferoffset+k];}
        for (k=52; k<76;k++) { DVSpoofBuffer3[k-52] = inbuffer[inbufferoffset+k];}

      }
      else if ((j>6) && (((j-6) % 16) != 0)) {// Ignore 6 header DIF blocks, as
                // Well as every 16th thereafter
                // (Audio DIF blocks) 

        for (k=4; k<80; k++) { 
          outbuffer[outbufferoffset] = inbuffer[inbufferoffset+k];
          checksumtemp += inbuffer[inbufferoffset+k];
          outbufferoffset++;
        } 
      }
    }
  }

  checksumtemp = checksumtemp << 8; // Mask out MSB 8-bits
  status->Checksum = checksumtemp >> 8;

  // Pick best 2 of three DVSpoofBuffers to be the real header info, and then parse it into
  // the FrameInfo struct
  if ( (BufferCompare(DVSpoofBuffer1, DVSpoofBuffer2, mask, 24, 0)) || 
         (BufferCompare(DVSpoofBuffer1, DVSpoofBuffer3, mask, 24, 0)) ) { // Pick 1st
    
    status->RealFrame = BufferCompare(DVSpoofBuffer1, DVSpoofKey, mask, 4,0);
    
    status->StoredChecksum = ((unsigned int)DVSpoofBuffer1[6] << 16) +
              ((unsigned int)DVSpoofBuffer1[5] << 8)  +
              ((unsigned int)DVSpoofBuffer1[4]);

    status->FileSize = ((_bigint)DVSpoofBuffer1[11] << 32) +
          ((_bigint)DVSpoofBuffer1[10] << 24) +  
              ((_bigint)DVSpoofBuffer1[9] << 16) +
              ((_bigint)DVSpoofBuffer1[8] << 8)  +
              ((_bigint)DVSpoofBuffer1[7]);

    status->FileCount = ((unsigned int)DVSpoofBuffer1[13] << 8) +
          ((unsigned int)DVSpoofBuffer1[12]);

    status->TotalFiles = ((unsigned int)DVSpoofBuffer1[15] << 8) +
          ((unsigned int)DVSpoofBuffer1[14]);

    status->FrameNum = ((unsigned int)DVSpoofBuffer1[18] <<  16) +
              ((unsigned int)DVSpoofBuffer1[17] << 8) +
          ((unsigned int)DVSpoofBuffer1[16]);

    status->TotalFrames = ((unsigned int)DVSpoofBuffer1[21] <<  16) +
              ((unsigned int)DVSpoofBuffer1[20] << 8) +
          ((unsigned int)DVSpoofBuffer1[19]);
  
  }
  else { // Pick 2nd
  
    status->RealFrame = BufferCompare(DVSpoofBuffer2, DVSpoofKey, mask, 4,0);
    
    status->StoredChecksum = ((unsigned int)DVSpoofBuffer2[6] << 16) +
              ((unsigned int)DVSpoofBuffer2[5] << 8)  +
              ((unsigned int)DVSpoofBuffer2[4]);

    status->FileSize = ((_bigint)DVSpoofBuffer2[11] << 32) +
          ((_bigint)DVSpoofBuffer2[10] << 24) +  
              ((_bigint)DVSpoofBuffer2[9] << 16) +
              ((_bigint)DVSpoofBuffer2[8] << 8)  +
              ((_bigint)DVSpoofBuffer2[7]);

    status->FileCount = ((unsigned int)DVSpoofBuffer2[13] << 8) +
          ((unsigned int)DVSpoofBuffer2[12]);

    status->TotalFiles = ((unsigned int)DVSpoofBuffer2[15] << 8) +
          ((unsigned int)DVSpoofBuffer2[14]);

    status->FrameNum = ((unsigned int)DVSpoofBuffer2[18] <<  16) +
              ((unsigned int)DVSpoofBuffer2[17] << 8) +
          ((unsigned int)DVSpoofBuffer2[16]);

    status->TotalFrames = ((unsigned int)DVSpoofBuffer2[21] <<  16) +
              ((unsigned int)DVSpoofBuffer2[20] << 8) +
          ((unsigned int)DVSpoofBuffer2[19]);
  
  }

  return 1;
}





//////////////////////////////////////////////////////////////////////////////////////////////////
// ParseStatus handles the different exception cases that can occur from corruption
// Cases that must be handled:
//  0)A Detected Frame 0 indicates a new output file (also an incremented filenumber)
//  1)Repeated Frame 0 (keep running checksum and only output 1 at most)
//  2)One or more frames are dropped (Pad 120kbytes)
//  3)One or more frames are dropped from the end of the input file (Pad < 120kbytes)
//  4)A frame is duplicated (suppress extra frames)
//////////////////////////////////////////////////////////////////////////////////////////////////
int ParseStatus (FrameInfoType * status, FrameInfoType * prevstatus) {

  // On the first Frame 0 detect, need to flush any remaining bytes
  // When writing, decrement status->FileSize
  // then, next status->FileSize should = prevstatus->Filesize otherwise bytes (frames) were dropped.

  if (status->FrameNum == 0) { // Special handling for Frame 0
//    if (status->FileCount != prevstatus->FileCount) { // New File in .avi
    if ( (prevstatus->FrameNum != 0) || (status->FileCount != prevstatus->FileCount) ) { // New File in .avi
      if (prevstatus->FileSize > 0) { // CASE 3
        status->Status = DROPPED_FRAMES + (prevstatus->TotalFrames-prevstatus->FrameNum);
        status->ZeroBytes = prevstatus->FileSize;
        status->RealBytes = 0; // Now we are decoding Frame0's which wont be locked in yet
      } 
      else {
        status->Status = REDUNDANT_FRAME0; // Dont lock in yet
        status-> FileSize = prevstatus-> FileSize;
        status->ZeroBytes = 0;
        status->RealBytes = 0;  
      }
      status-> Frame0Count = 0; // Reset Framecount 
    }
    else {
      if ( (status->Checksum == prevstatus->Checksum) && (prevstatus->Frame0Count < 3) ) { // CASE 1
        status->Frame0Count = prevstatus->Frame0Count + 1;
        if (status->Frame0Count == 3) { // Lock in Frame0
          status->Frame0Count = 4;
          status->Status = LOCK_FRAME0;
          status->ZeroBytes = 0;
          if (status-> FrameNum == status-> TotalFrames) { // Last Frame of this File 
            status-> RealBytes = status-> FileSize;
            status-> FileSize = 0; 
          }
          else {
            status-> RealBytes = 102524-(14*76); // Less Frame0 headers for filename,etc
            status-> FileSize -= 102524-(14*76);  
          }

        }
        else {
          status->Status = REDUNDANT_FRAME0; // Dont lock in yet
          status-> FileSize = prevstatus-> FileSize;
          status->ZeroBytes = 0;
          status->RealBytes = 0;
        }     
      }
      else { // Either checksum doesnt match or framecount = 4
        if (prevstatus->Frame0Count < 4) { // If Its not already locked in
          status->Frame0Count = 0;
          status->Status = RESET_FRAME0;
          status->ZeroBytes = 0;
          status->RealBytes = 0;
        }
        else { // Who cares if the checksum doesnt match, Frame0 is already locked in
          status->Status = REDUNDANT_FRAME0;
          status-> FileSize = prevstatus-> FileSize;
          status->ZeroBytes = 0;
          status->RealBytes = 0;
        }
      }
    }
  }
  else { // Non Frame0
    status->Frame0Count = 0;
    if (status->FrameNum == prevstatus->FrameNum) { // CASE 4, Duplicate Frame
      status->Status = DUPLICATE_FRAME;
      status->ZeroBytes = 0;
      status->RealBytes = 0;
      
      if (status->FrameNum == status->TotalFrames) { // Last Frame of this File
        status->FileSize = 0;
      }
      else {
        status->FileSize -= 102524;
      }
    }
    else if (status->FrameNum > (prevstatus->FrameNum + 1)) { // CASE 2, Dropped Frames
      status->Status = DROPPED_FRAMES + (status->FrameNum - prevstatus->FrameNum - 1);
      status->ZeroBytes = (102524 * (status->FrameNum - prevstatus->FrameNum - 1)); 

      if (status->FrameNum == status->TotalFrames) { // Last Frame of this File
        status->RealBytes = status->FileSize;
        status->FileSize = 0;
      }
      else {
        status->RealBytes = 102524;
        status->FileSize -= 102524;
      }
    }

    else { // Usual Case, No problems
      status->Status = EXPECTED_FRAME;
      status->ZeroBytes = 0;

      // TotalFrames is an count of the #AVI frames in the file, including
      // both PARITY and DATA frames.  We no longer use TotalFrames to
      // identify the end of a datafile.
      //if (status->FrameNum == status->TotalFrames) { // Last Frame of this File
      if (status->FileSize < 102524 ) { // Last Frame of this File
        status->RealBytes = status->FileSize;
        status->FileSize = 0; 
      }
      else {
        status->RealBytes = 102524;
        status->FileSize -= 102524;
      }
    }
  } 

  return 1;
}


// auto-recursive "mkdir" command
// automatically converts a hierarchical path into its base-elements,
// makes each sub/directory from base to subpath.
int
mkdir_recursive( const char pathname[] )
{
  char currentpath[ 255 ], dircommand[ 255 + 8 ];
  int i,x;

  x = 0;
  if ( pathname[ 0 ] == '\\' )
    ++x; // skip the first backslash

  // the following loop performs this operation on pathname="\c\d\e\f"
  //   1) mkdir "\c"
  //   2) mkdir "\c\d"
  //   3) mkdir "\c\d\e"
  //   4) mkdir "\c\d\e\f"

  while ( pathname[ x ] != '\0' )
  {
    // advance 'x' to point to the next backslash (or null-char)
    while ( pathname[ x ] != '\0' && pathname[ x ] != '\\' )
      ++x; // increment marker 'x'

    // copy the string manually, up to (but not including) the marker 'x' 
    for ( i = 0; i < x; ++i )
      currentpath[ i ] = pathname[ i ];
    currentpath[ i ] = '\0'; // add string terminator

    // construct string : mkdir "<currentpath>" 
    sprintf( dircommand, "mkdir \"%s\"", currentpath );
    //strcat(dircommand, " 2>SPOOFDIR"); // MSVC 6 + Win98SE doesn't like
    printf("\nmkdir_recursive( '%s' ) : '%s'", pathname, dircommand );
    system( dircommand );

    // now advance the marker past the current '\'
    if ( pathname[ x ] != '\0' )
      ++x;
  }
  
  return 0;
}


// NOTE: THIS TURNS OFF GLOBBING SUPPORT UNDER MINGW32 COMPILER. GLOBBING MUST BE TURNED OFF!
int _CRT_glob = 0; 

int
dump_eccblock( const FrameInfoType pfibuffer[] )
{

  char outfilename[255];
  char outfilepath[255];
  //char infilename[255];
  //char startfilename[255];
  //char dircommand[255];

  long totalframes = 0, totaldroppedframes = 0, longestdroppedframesequence = 0,
       totalduplicatedframes = 0, totalcorruptedframes = 0, 
       totalunrecognizableframes = 0;
  _bigint totalbytes = 0;

  //int recurse = 0;
  int inplace = 0;
  int SLEEPTIME = 0;
  int startfilefound = 1;
  int poutfilename, poutfilepath;

  FrameInfoType prevframestatus = {0}, framestatus = {0};
  int ecc_k;
  int i,j;
  eccword  *ECCdatabuffer;   // pointer to the decoded data of the ECC_block 
  unsigned char *DataBuffer; // alias for ECCdatabuffer
  unsigned int ECCchecksumtemp = 0;
  int rc; // function return code

  framestatus.Frame0Count = 0;
  framestatus.FrameNum = -1;

  printf("\ndump_eccblock():");

  for ( ecc_k = 0; ecc_k < ECC_K; ++ecc_k )
  // Call ReadDVFrame to parse .avi file(s) and return the next DV frame
  { 

    prevframestatus.Frame0Count = framestatus.Frame0Count;
    prevframestatus.FileCount = framestatus.FileCount;
    prevframestatus.TotalFiles = framestatus.TotalFiles;
    prevframestatus.FrameNum = framestatus.FrameNum;
    prevframestatus.TotalFrames = framestatus.TotalFrames;
    prevframestatus.FileSize = framestatus.FileSize;
    prevframestatus.Checksum = framestatus.Checksum;
    prevframestatus.StoredChecksum = framestatus.Checksum;
    prevframestatus.Status = framestatus.Status;
  
    framestatus = pfibuffer[ ecc_k ];
    rc = ecc_decoder_get_out_buffer( &g_eccDecoderState, &ECCdatabuffer );
    DataBuffer = (unsigned char *) ECCdatabuffer;

    // If a valid header was found, the frame is Real and should be processed
    if (framestatus.RealFrame) {

      // Determine if frames have been dropped or duplicated, etc
      //error = ParseStatus(&framestatus, &prevframestatus);
    
      // If we've locked in a Frame0, close the old output file and open a new one
      if (framestatus.Status == LOCK_FRAME0) {
        if (g_outfile != NULL) { fclose(g_outfile); }

        // Get filepath and filename from First 228 bytes (3 DIF blocks) of Frame0
        poutfilename = 0;
        poutfilepath = 0;
        outfilepath[0] = '\0';
        for (i=0;i<228;i++) {
          if (DataBuffer[i] == '\\') {
            poutfilename = 0;
            poutfilepath = i+1; // 0->i is path, i+1->228 is name
          }
          else {
            poutfilename++;
          }
        }
        memcpy(outfilepath,DataBuffer,poutfilepath);
        outfilepath[poutfilepath]='\0';
        if ((g_recurse) && (outfilepath[0] != '\0') ){ // Make dir and use full path of file
          mkdir_recursive( outfilepath );  // perform a recursive 'mkdir'
          //strcpy(dircommand,"mkdir ");
          //strcat(dircommand,outfilepath);
          //strcat(dircommand, " 2>SPOOFDIR"); // MSVC 6 + Win98SE doesn't like
          //system(dircommand);
          //remove("SPOOFDIR");
          memcpy(outfilename,DataBuffer,228);
        }
        else { // create files in current directory 
          memcpy(outfilename,DataBuffer+poutfilepath,228-poutfilepath);
        }
        
        // case-insensitive
        if ((!strcmpi(outfilename,g_startfilename)) || (g_startfilefound)) {
//      if ((!strcmp(outfilename,g_startfilename)) || (g_startfilefound)) {
          g_startfilefound = 1;
          g_outfile = fopen (outfilename, "wb"); 
          if (g_outfile == NULL) 
          { 
            printf ("Error opening output file '%s'!\n", outfilename );
            return (-1);
          }
        }
      }

      // Print Status
      // case-insensitive
      if ((!strcmpi(outfilename,g_startfilename)) || (startfilefound)) { // We've found the startfile or we're not using one
//    if ((!strcmp(outfilename,g_startfilename)) || (startfilefound)) { // We've found the startfile or we're not using one
    
        switch (framestatus.Status) { 

          case LOCK_FRAME0:
            totalframes++;

            printf ("\nWriting to %s (%d Frames): 0",outfilename,framestatus.TotalFrames+1);
            fprintf (g_logfile, "\nWriting to %s (%d Frames): 0",outfilename,framestatus.TotalFrames+1);
            break;
  
          case REDUNDANT_FRAME0:
            // this case should NEVER happen
            printf ("\ndump_eccblock() : CATASTROPHIC ERROR, REDUNDANT_FRAME0!" );
            fprintf (g_logfile, "\ndump_eccblock() : CATASTROPHIC ERROR, REDUNDANT_FRAME0!" );
            break;
      
          case RESET_FRAME0:
            printf ("!");
            fprintf (g_logfile,"!");
            break;
      
          case EXPECTED_FRAME:
            totalframes++;
//            fprintf (g_logfile,"\nGot expected # %d", framestatus.FrameNum);
            if (framestatus.Checksum != framestatus.StoredChecksum) { // Corrupted
              printf ("*"); 
              fprintf (g_logfile,"*");
              totalcorruptedframes++;
            }
            else if ((framestatus.FrameNum == framestatus.TotalFrames) && (framestatus.FrameNum != prevframestatus.FrameNum) ) {
              printf ("%d",framestatus.FrameNum);
              fprintf (g_logfile,"%d",framestatus.FrameNum);
            }
            else if (((framestatus.FrameNum % 10) == 0) && (framestatus.FrameNum != 0) && (framestatus.FrameNum != prevframestatus.FrameNum)) {
              printf ("%d",framestatus.FrameNum);
              fprintf (g_logfile,"%d",framestatus.FrameNum);
            }
            else {
              printf (".");
              fprintf (g_logfile,".");
            }
            break;
    
          case DUPLICATE_FRAME:
            totalduplicatedframes++;    

            printf ("+");
            fprintf (g_logfile,"+");
            break;

          default: // DROPPED_FRAMES NOTE: If there are more than 0x7FFFFFFF dropped frames everything will screw up royally!
            fprintf( g_logfile, "\ndump_eccblock() : Dropped_frame %d", framestatus.Status );
            totaldroppedframes+= framestatus.Status;
            totalframes = totalframes + framestatus.Status + 1;
            if (framestatus.Status > longestdroppedframesequence) {
              longestdroppedframesequence = framestatus.Status;
            }
            for (j=0;j<framestatus.Status;j++) {
              printf ("-");
              fprintf (g_logfile,"-");
            }
            if (framestatus.Checksum != framestatus.StoredChecksum) { // Corrupted
              printf ("*"); 
              fprintf (g_logfile,"*");
              totalcorruptedframes++;
            }
            else if ((framestatus.FrameNum == framestatus.TotalFrames) && (framestatus.FrameNum != prevframestatus.FrameNum) ) {
              printf ("%d",framestatus.FrameNum);
              fprintf (g_logfile,"%d",framestatus.FrameNum);
            }
            else if (((framestatus.FrameNum % 10) == 0) && (framestatus.FrameNum != 0) && (framestatus.FrameNum != prevframestatus.FrameNum)) {
              printf ("%d",framestatus.FrameNum);
              fprintf (g_logfile,"%d",framestatus.FrameNum);
            }
            else if (framestatus.FrameNum == 0) {
              // print nothing
            }
            else {
              printf (".");
              fprintf (g_logfile,".");
            }
            break;
    
        }

        // ParseStatus tells us how many bytes of padding and how many bytes of payload to write out
        if ( (framestatus.ZeroBytes > 0) && (g_outfile != NULL) ){
          printf("\nZerobytes = %d, RealBytes = %d", framestatus.ZeroBytes, framestatus.RealBytes );
/*          for (i=0; i<framestatus.ZeroBytes;i++) {
            fputc(0x00, g_outfile);
          }*/
        }
        totalbytes += (double)framestatus.ZeroBytes;

        if (framestatus.Status == LOCK_FRAME0) { // Skip filename headers for Frame0
          if ((framestatus.RealBytes > 0) && (g_outfile != NULL) ){
            // recompute checksum of outgoing data
            ECCchecksumtemp  = 0;
            for ( i = 0; i < framestatus.RealBytes; ++i )
              ECCchecksumtemp  += DataBuffer[i+(14*76)];
            if ( (framestatus.Checksum != ECCchecksumtemp) && 
              (framestatus.StoredChecksum == ECCchecksumtemp) )
              --totalcorruptedframes;

            fwrite(DataBuffer+(14*76), framestatus.RealBytes, 1, g_outfile);
//            printf ("\nWrite frame0 #%d, %d bytes",framestatus.FrameNum, framestatus.RealBytes);
          } 
        }
        else {
          if ((framestatus.RealBytes > 0) && (g_outfile != NULL) ){
            // recompute checksum of outgoing data
            ECCchecksumtemp  = 0;
            for ( i = 0; i < framestatus.RealBytes; ++i )
              ECCchecksumtemp  += DataBuffer[i];
              if ( (framestatus.Checksum != ECCchecksumtemp) && 
                (framestatus.StoredChecksum == ECCchecksumtemp) )
                --totalcorruptedframes;

            fwrite(DataBuffer, framestatus.RealBytes, 1, g_outfile);
//            printf ("\nWrite frame #%d, %d bytes",framestatus.FrameNum, framestatus.RealBytes);
          } 
        }
        totalbytes += framestatus.RealBytes;


        // Go to sleep (to reduce CPU usage and get closer to 'real-time' performance for ATSC streams
        if (framestatus.FrameNum != 0) {
//            sleep(SLEEPTIME);
            ;
        }
      }
    }
    else { // Unrecognizable Frame, ignore it
      totalunrecognizableframes++;

      printf ("\ndump_eccblock() : UH OH, unrecognized frame!");
      fprintf (g_logfile,"X");
    }

  } // end while

  // currently, statistics are not returned to the caller;
  // (information like totalcorruptedframes, totalunrecognizableframes, etc.)

  return 0;
} 


int
add2eccqueue( const unsigned char DataBuffer[], 
  const FrameInfoType *framestatus, const long dataframecount )
{
  int rc;
  eccword *ecc_DataBuffer = g_eccDecoderState.pReceivedPlanes[ g_ecc_n ];
  eccword *ecc_DataBuffer2 = NULL; 

//  printf("\nadd2queue( g_ecc_n = %d )", g_ecc_n );
  rc = ecc_decoder_get_rcv_buffer( &g_eccDecoderState, &ecc_DataBuffer2 );
  if ( ecc_DataBuffer2 != ecc_DataBuffer )
  {
    printf("\nadd2queue() : CATASTROPHIC ERROR!" );
    exit(0);
  }

  memcpy( (void *)ecc_DataBuffer, (const void *) DataBuffer, sizeof(eccword) * ECC_Z );

  // update ECC-decoder's syndrome-buffer
  rc = ecc_decoder_syn1( &g_eccDecoderState, ecc_DataBuffer );
  g_ecc_fibuffer[ g_ecc_n ] = *framestatus;
  //g_ecc_fibuffer[ g_ecc_n ].TotalFrames = dataframecount; // 
  ++g_ecc_n;

  if ( g_ecc_n == ECC_N ) // one full-ecc block yet?
  {
    g_ecc_n = 0; // reset counter
    // g_eccEnable=1 : do correction
    // g_eccEnable=0 : no correction
    // ecc_decoder_correct() eats a lot of compute cycles!
    rc = ecc_decoder_correct( &g_eccDecoderState, g_eccEnable );
    dump_eccblock( g_ecc_fibuffer );
  }

  return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
//Main
//////////////////////////////////////////////////////////////////////////////////////////////////////
int main(int argc, char * argv[]) {

  int error,i;
  int starttime,endtime,totalsec; 
  _bigint totalbytes;
  float totalmin,speed;
  long totalframes, totaldroppedframes, longestdroppedframesequence, totalduplicatedframes, totalcorruptedframes, totalunrecognizableframes;
  long dataframes;
  FILE * infile = NULL;
//  FILE * outfile = NULL;
  //char outfilename[255];
//  char outfilepath[255];
  char infilename[255];
  //char g_startfilename[255];
  //int g_startfilefound = 1;
  //char dircommand[255];

  int inplace = 0;
  int SLEEPTIME = 0;
//  int poutfilename, poutfilepath;

//  FILE * test; 
  unsigned char *DVFrameBuffer;
  unsigned char *DataBuffer;

  FrameInfoType framestatus = {0};  // needs to be initialized to 0!
  FrameInfoType prevframestatus = {0}; // needs to be initialized to 0!
  int rc; // function return code

  g_outfile = NULL;
  g_eccEnable = 1; // default, enable ECC-error correction
  g_startfilefound = 1; // default, turn on output writer on first file
  g_recurse = 0; // flag, 1 = recurse subdirectories

  // allocate memory
  g_ecc_fibuffer = (FrameInfoType *) malloc( sizeof( FrameInfoType ) * ECC_N );
  DVFrameBuffer = (unsigned char *) malloc( sizeof( unsigned char) * 120000 );
  DataBuffer = (unsigned char *) malloc( sizeof( unsigned char) * 102524 );
  if ( g_ecc_fibuffer == NULL || DVFrameBuffer == NULL || DataBuffer == NULL )
  {
    printf("\nmain() : Error, malloc() failed." );
    exit( EXIT_FAILURE );
  }

  // initialize the ECC_decoder
  for ( i = 0; i < ECC_N; ++i )
    g_eccDecoderState.pReceivedPlanes[i] = 0; // initializing for first time
  for ( i = 0; i < ECC_NMK; ++i )
    g_eccDecoderState.pSyndromePlanes[i] = 0; // initializing for first time

  ecc_decoder_init( &g_eccDecoderState );

  if (argc == 1) { 
    printf ("DVdec v2.0 BETA - By Andy Kromkamp\n");
    printf ("  (Hamming-code ECC by Royce Liao)\n");
    printf ("Usage: DVdec [-throttle #msec] [-startfile filename] [-inplace] [-recurse] <First .avi file>\n\n");
//    printf ("  -throttle: Slows down the decoding to reduce CPU usage. Defaults to 0\n");
//        printf ("             (no delay), higher numbers are slower.\n");
    printf ("  -throttle: NOT SUPPORTED\n");
    printf ("  -startfile: Specifies which file to start decoding at. If the -recurse\n");
    printf ("              parameter is also specified, the complete path of the file\n");
    printf ("              must be specified also.\n");
    printf ("  -inplace: Deletes the input .avi files as they are used up (to save drive\n");
    printf ("            space)\n");
    printf ("  -nocorrect: disable ECC-error correction (ignore parity information)\n");
    printf ("  -recurse: Recreate files with complete filepaths. Without this parameter,\n");
    printf ("            files are extracted to the current directory.\n\n");
                                                                                                                                                                                                                                                                                                                                         

    printf ("Example: DVdec SCENE0001.avi\n");
    printf ("Example: DVdec -startfile data.ts.0005 SCENE0002.avi\n");
    printf ("Example: DVdec -recurse -startfile \\directory1\\file1 SCENE0003.avi\n");
    printf ("Example: DVdec -throttle 30 DV0003.avi\n");
    printf ("Example: DVdec -inplace scene0001.avi\n");
    printf ("Example: DVdec -recurse scene0001.avi\n");
    printf ("NOTE: PAL Camcorders are not supported at this time.\n");
    return (-1); 
  } 

  for (i=1; i< argc; i++) {
    if (!strcmp(argv[i],"-inplace")) {
      inplace = 1;          
    }
    else if (!strcmp(argv[i],"-recurse")) {
      g_recurse = 1;
    }
    else if (!strcmp(argv[i],"-throttle")) {
      SLEEPTIME = atoi(argv[i+1]);
      i++;
    }
    else if (!strcmp(argv[i],"-startfile")) {
      strcpy(g_startfilename, argv[i+1]);  
      g_startfilefound = 0;
      i++;
    }
    else if (!strcmp(argv[i],"-nocorrect")) {
      g_eccEnable=0; // disable ecc-correction
      i++;
    }   else { // Must be name of first input .avi file
      strcpy(infilename,argv[i]);
      i = argc;
    }
  }

  framestatus.Frame0Count = 0;
  framestatus.FrameNum = -1;

  totalbytes = 0;
  dataframes = 0;
  totalframes = 0;
  totaldroppedframes = 0;
  longestdroppedframesequence = 0;
  totalduplicatedframes = 0;
  totalcorruptedframes = 0;
  totalunrecognizableframes = 0;

  starttime = time(NULL);

  g_logfile = fopen ("DVdeclog.txt", "w");
  if (g_logfile == NULL) { printf ("Error opening file!\n"); return (-1); }
  g_ecc_n = 0;
  
  // Call ReadDVFrame to parse .avi file(s) and return the next DV frame
  while (ReadDVFrame(DVFrameBuffer,&infile,infilename,inplace,&g_logfile)) {
 
    prevframestatus.Frame0Count = framestatus.Frame0Count;
    prevframestatus.FileCount = framestatus.FileCount;
    prevframestatus.TotalFiles = framestatus.TotalFiles;
    prevframestatus.FrameNum = framestatus.FrameNum;
    prevframestatus.TotalFrames = framestatus.TotalFrames;
    prevframestatus.FileSize = framestatus.FileSize;
    prevframestatus.Checksum = framestatus.Checksum;
    prevframestatus.StoredChecksum = framestatus.Checksum;
    prevframestatus.Status = framestatus.Status;
  
    // Split DV frame into Header data (Video DIF block 0) and Payload Data
    error = SplitDVFrame(DVFrameBuffer,DataBuffer,&framestatus);

    // If a valid header was found, the frame is Real and should be processed
    if (framestatus.RealFrame) {

      // Determine if frames have been dropped or duplicated, etc
      error = ParseStatus(&framestatus, &prevframestatus);
   
      // Print Status
//      if ((!strcmp(outfilename,startfilename)) || (startfilefound)) { // We've found the startfile or we're not using one
        if ( 1 ) { // unconditionally execute this if-branch
    
        switch (framestatus.Status) { 

          case LOCK_FRAME0:
            // push buffer into ECC_decoder's queue.
            if (framestatus.Checksum != framestatus.StoredChecksum) { // Corrupted
              totalcorruptedframes++;
            }
            rc = add2eccqueue( DataBuffer, &framestatus, dataframes );
            ++totalframes;
            ++dataframes;
            break;

          case REDUNDANT_FRAME0:
            // ignore
            break;
  
          case RESET_FRAME0:
            // don't know what to do here, probably ignore
            printf("\nmain() : RESET_FRAME0 @ g_ecc_n=%d", g_ecc_n );
            break;
      
          case EXPECTED_FRAME:
            // push buffer onto stack.
            if (framestatus.Checksum != framestatus.StoredChecksum) { // Corrupted
              totalcorruptedframes++;
            }
            rc = add2eccqueue( DataBuffer, &framestatus, dataframes );
            ++totalframes;
            if ( g_ecc_n < ECC_K )
              ++dataframes;
            break;
    
          case DUPLICATE_FRAME:
            //printf("\nmain() : Duplicate Frame #%d", totalduplicatedframes );
            totalduplicatedframes++;    
            break;

          default: // DROPPED_FRAMES NOTE: If there are more than 0x7FFFFFFF dropped frames everything will screw up royally!
            fprintf( g_logfile, "\nmain() : Dropped_frame %d", framestatus.Status );

            // dump as many frames as were dropped
            // push buffer onto stack.
            for ( i = 0; i < (framestatus.Status + 1); ++i )
            {
              rc = add2eccqueue( DataBuffer, &framestatus, dataframes );
              if ( g_ecc_n < ECC_K )
                ++dataframes;
            }

            totaldroppedframes+= framestatus.Status;
            totalframes = totalframes + framestatus.Status + 1;
//            totalframes = totalframes + framestatus.Status + 1;
            if (framestatus.Status > longestdroppedframesequence) {
              longestdroppedframesequence = framestatus.Status;
            }

            break;
    
        }

        totalbytes += (double)framestatus.ZeroBytes;
        totalbytes += framestatus.RealBytes;


        // Go to sleep (to reduce CPU usage and get closer to 'real-time' performance for ATSC streams
        if (framestatus.FrameNum != 0) {
//            sleep(SLEEPTIME);
            ;
        }
      }
    }
    else { // Unrecognizable Frame, ignore it
      totalunrecognizableframes++;

      framestatus.Frame0Count = prevframestatus.Frame0Count;
      framestatus.FileCount = prevframestatus.FileCount;
      framestatus.TotalFiles = prevframestatus.TotalFiles;
      framestatus.FrameNum = prevframestatus.FrameNum;
      framestatus.TotalFrames = prevframestatus.TotalFrames;
      framestatus.FileSize = prevframestatus.FileSize;
      framestatus.Checksum = prevframestatus.Checksum;
      framestatus.StoredChecksum = prevframestatus.Checksum;
      framestatus.Status = prevframestatus.Status;

      printf ("X");
      fprintf (g_logfile,"X");
    }

  } // end while

  if ( g_ecc_n != 0 )
  {
    printf( "\nmain() : g_ecc_n == %d (should be 0), uh oh...", g_ecc_n );
  }

  if (totalframes == 0) {
    printf ("No Input Files Found!\n");
    return 0;
  }

  endtime = time(NULL);
  totalsec = endtime - starttime; // in seconds
  totalmin = totalsec / 60; // in minutes
  if (totalsec > 0) {
    speed = 0.0 + (totalbytes / totalmin);
  }
  else {
    speed = 0.0;
  }
  speed = speed / 150000000.0;

  printf ("\n\n\nStatistics\n----------\n");
  printf ("Total Frames decoded(.): %d\n", totalframes);
  printf ("Total Dropped Frames(-): %d\n", totaldroppedframes);
  printf ("(Longest # of Consecutive Frame Drops: %d)\n", longestdroppedframesequence);
  printf ("Total Duplicated Frames(+): %d\n", totalduplicatedframes);
  printf ("Total Corrupted Frames(*): %d\n", totalcorruptedframes);
  printf ("Total Unrecognizable Frames(X): %d\n",totalunrecognizableframes);
  printf ("NOTE : THESE STATISTICS APPLY TO THE RAW AVIFILE BEFORE ECC_CORRECTION!\n");

  printf ("\n%d Bytes decoded in ",totalbytes);
  printf ("%3.2f minutes\n",totalmin);
  printf ("(ATSC data is ~150Mbytes/minute, so\n");
  printf ("performance was %3.2f X ATSC real-time)\n",speed);

  fprintf (g_logfile,"\n\n\nStatistics\n----------\n");
  fprintf (g_logfile,"Total Frames decoded(.): %d\n", totalframes);
  fprintf (g_logfile,"Total Dropped Frames(-): %d\n", totaldroppedframes);
  fprintf (g_logfile,"(Longest # of Consecutive Frame Drops: %d)\n", longestdroppedframesequence);
  fprintf (g_logfile,"Total Duplicated Frames(+): %d\n", totalduplicatedframes);
  fprintf (g_logfile,"Total Corrupted Frames(*): %d\n", totalcorruptedframes);
  fprintf (g_logfile,"Total Unrecognizable Frames(X): %d\n",totalunrecognizableframes);
  fprintf (g_logfile,"\n%d Bytes decoded in ",totalbytes);
  fprintf (g_logfile,"%3.2f minutes\n",totalmin);
  fprintf (g_logfile,"(ATSC data is ~150Mbytes/minute, so\n");
  fprintf (g_logfile,"performance was %3.2f X ATSC real-time)\n",speed);

  fclose (g_logfile);

  return 1;
}

/*

 
  // Call ReadDVFrame to parse .avi file(s) and return the next DV frame
  while (ReadDVFrame(DVFrameBuffer,&infile,infilename,inplace,&g_logfile)) {

    prevframestatus.Frame0Count = framestatus.Frame0Count;
    prevframestatus.FileCount = framestatus.FileCount;
    prevframestatus.TotalFiles = framestatus.TotalFiles;
    prevframestatus.FrameNum = framestatus.FrameNum;
    prevframestatus.TotalFrames = framestatus.TotalFrames;
    prevframestatus.FileSize = framestatus.FileSize;
    prevframestatus.Checksum = framestatus.Checksum;
    prevframestatus.StoredChecksum = framestatus.Checksum;
    prevframestatus.Status = framestatus.Status;
  
    // Split DV frame into Header data (Video DIF block 0) and Payload Data
    error = SplitDVFrame(DVFrameBuffer,DataBuffer,&framestatus);

    // If a valid header was found, the frame is Real and should be processed
    if (framestatus.RealFrame) {

      // Determine if frames have been dropped or duplicated, etc
      error = ParseStatus(&framestatus, &prevframestatus);
    
      // If we've locked in a Frame0, close the old output file and open a new one
      if (framestatus.Status == LOCK_FRAME0) {
        if (outfile != NULL) { fclose(outfile); }

        // Get filepath and filename from First 228 bytes (3 DIF blocks) of Frame0
        poutfilename = 0;
        poutfilepath = 0;
        outfilepath[0] = '\0';
        for (i=0;i<228;i++) {
          if (DataBuffer[i] == '\\') {
            poutfilename = 0;
            poutfilepath = i+1; // 0->i is path, i+1->228 is name
          }
          else {
            poutfilename++;
          }
        }
        memcpy(outfilepath,DataBuffer,poutfilepath);
        outfilepath[poutfilepath]='\0';
        if ((recurse) && (outfilepath[0] != '\0') ){ // Make dir and use full path of file
          strcpy(dircommand,"mkdir ");
          strcat(dircommand,outfilepath);
          strcat(dircommand, " 2>SPOOFDIR");
          system(dircommand);
          remove("SPOOFDIR");
          memcpy(outfilename,DataBuffer,228);
        }
        else { // create files in current directory 
          memcpy(outfilename,DataBuffer+poutfilepath,228-poutfilepath);
        }
        
        if ((!strcmp(outfilename,startfilename)) || (startfilefound)) {
          startfilefound = 1;
          outfile = fopen (outfilename, "wb"); 
          if (outfile == NULL) { printf ("Error opening output file!\n"); return (-1); }
        }
      }

      // Print Status
      if ((!strcmp(outfilename,startfilename)) || (startfilefound)) { // We've found the startfile or we're not using one
    
        switch (framestatus.Status) { 

          case LOCK_FRAME0:
            totalframes++;

            printf ("\nWriting to %s (%d Frames): 0",outfilename,framestatus.TotalFrames+1);
            fprintf (g_logfile, "\nWriting to %s (%d Frames): 0",outfilename,framestatus.TotalFrames+1);
            break;
  
          case REDUNDANT_FRAME0:
            break;
      
          case RESET_FRAME0:
            printf ("!");
            fprintf (g_logfile,"!");
            break;
      
          case EXPECTED_FRAME:
            totalframes++;
            if (framestatus.Checksum != framestatus.StoredChecksum) { // Corrupted
              printf ("*"); 
              fprintf (g_logfile,"*");
              totalcorruptedframes++;
            }
            else if ((framestatus.FrameNum == framestatus.TotalFrames) && (framestatus.FrameNum != prevframestatus.FrameNum) ) {
              printf ("%d",framestatus.FrameNum);
              fprintf (g_logfile,"%d",framestatus.FrameNum);
            }
            else if (((framestatus.FrameNum % 10) == 0) && (framestatus.FrameNum != 0) && (framestatus.FrameNum != prevframestatus.FrameNum)) {
              printf ("%d",framestatus.FrameNum);
              fprintf (g_logfile,"%d",framestatus.FrameNum);
            }
            else {
              printf (".");
              fprintf (g_logfile,".");
            }
            break;
    
          case DUPLICATE_FRAME:
            totalduplicatedframes++;    

            printf ("+");
            fprintf (g_logfile,"+");
            break;

          default: // DROPPED_FRAMES NOTE: If there are more than 0x7FFFFFFF dropped frames everything will screw up royally!
            totaldroppedframes+= framestatus.Status;
            totalframes = totalframes + framestatus.Status + 1;
            if (framestatus.Status > longestdroppedframesequence) {
              longestdroppedframesequence = framestatus.Status;
            }
            for (j=0;j<framestatus.Status;j++) {
              printf ("-");
              fprintf (g_logfile,"-");
            }
            if (framestatus.Checksum != framestatus.StoredChecksum) { // Corrupted
              printf ("*"); 
              fprintf (g_logfile,"*");
              totalcorruptedframes++;
            }
            else if ((framestatus.FrameNum == framestatus.TotalFrames) && (framestatus.FrameNum != prevframestatus.FrameNum) ) {
              printf ("%d",framestatus.FrameNum);
              fprintf (g_logfile,"%d",framestatus.FrameNum);
            }
            else if (((framestatus.FrameNum % 10) == 0) && (framestatus.FrameNum != 0) && (framestatus.FrameNum != prevframestatus.FrameNum)) {
              printf ("%d",framestatus.FrameNum);
              fprintf (g_logfile,"%d",framestatus.FrameNum);
            }
            else if (framestatus.FrameNum == 0) {
              // print nothing
            }
            else {
              printf (".");
              fprintf (g_logfile,".");
            }
            break;
    
        }

        // ParseStatus tells us how many bytes of padding and how many bytes of payload to write out
        if ( (framestatus.ZeroBytes > 0) && (outfile != NULL) ){
          for (i=0; i<framestatus.ZeroBytes;i++) {
            fputc(0x00, outfile);
          }
        }
        totalbytes += (double)framestatus.ZeroBytes;

        if (framestatus.Status == LOCK_FRAME0) { // Skip filename headers for Frame0
          if ((framestatus.RealBytes > 0) && (outfile != NULL) ){
            fwrite(DataBuffer+(14*76), framestatus.RealBytes, 1, outfile);
          } 
        }
        else {
          if ((framestatus.RealBytes > 0) && (outfile != NULL) ){
            fwrite(DataBuffer, framestatus.RealBytes, 1, outfile);
          } 
        }
        totalbytes += framestatus.RealBytes;


        // Go to sleep (to reduce CPU usage and get closer to 'real-time' performance for ATSC streams
        if (framestatus.FrameNum != 0) {
//            sleep(SLEEPTIME);
            ;
        }
      }
    }
    else { // Unrecognizable Frame, ignore it
      totalunrecognizableframes++;

      framestatus.Frame0Count = prevframestatus.Frame0Count;
      framestatus.FileCount = prevframestatus.FileCount;
      framestatus.TotalFiles = prevframestatus.TotalFiles;
      framestatus.FrameNum = prevframestatus.FrameNum;
      framestatus.TotalFrames = prevframestatus.TotalFrames;
      framestatus.FileSize = prevframestatus.FileSize;
      framestatus.Checksum = prevframestatus.Checksum;
      framestatus.StoredChecksum = prevframestatus.Checksum;
      framestatus.Status = prevframestatus.Status;

      printf ("X");
      fprintf (g_logfile,"X");
    }

  } // end while


*/