// DVenc
// By: Andy Kromkamp kromkamp@my-deja.com
//
// THIS SOURCE CODE IS GIVEN AWAY FREELY, WITH NO GUARANTEES EXPLICIT OR IMPLIED.

/////////////////////////////////////////////////////////////////////////////
//  DVSPOOF(ECC_hamming) v0.1 09/29/2001 liaor@iname.com
// 
//    This version of DVSPOOF has been modified to support forward-ECC.
//  It creates AVI-files with ECC-information (this cannot be turned off),
//  and is therefore NOT compatible with the original DVSPOOF.  
// 
//  Ecc-encoder
//  -----------
//    The ECC-encoder intercepts outgoing binary-data (before they are
//  packed into AVIframes), calculates parity-data, then re-inserts the
//  parity-check codes into the AVIfile.
//
//    ReadData() -> ECC_encoder -> FillDVFrame() -> WriteDV() -> *.AVI
//
//    ReadData() sends "payload data" to the ECC_encoder.  The ECC-encoder
//  calculates parity information,and organizes the data+parity into one
//  "ecc block."  The entire ecc_block (64 data frames + 8 parity frames)
//  is then sent to AVIframe_packer ( FillDVFrame() ).  Finally the 
//  WriteDV() dumps the AVIframes to the AVIfile.
//    The actual code deviates somewhat from to the above behavior.
//  First of all, dvspoof places some additional header-information into
//  the first-frame of the AVI-file.  This info occupies the payload region,
//  meaning that the 1st AVIframe contains less user-data than non-first
//  AVIframes.  Since the header-info is stored in the payload region, 
//  the ECC_encoder protects it.  Secondly, the 1st AVIframe of any encoded
//  file is repeated several times (30).  WriteDV() repeats the AVIframe.
//  Since the repeat process occurs *AFTER* the ECC_encoding process, the 
//  first ECCblock in the ECC_file appears 'stretched.'
//    Also of note, the ECC_encoder forces the entire program to 
//  work in units of ECC-blocks.  This has an impact on the size of an
//  AVIfile, whose length is now constrained to a multiple of ECC_blocks.
//  (6400KB, 12800KB, 19200KB, etc.)  The constraint applies to the
//  output AVIfiles, not to input files.
//  
//
//  Terms
//  -----
//    AVIfile - Output unit for dvspoof.exe
//    AVIframe - one frame (audio + video) 
//        Data is stored only in the video-portion of the AVIframe.
//        The audio-portion is unused.
//    ECCblock - Output unit for the ECC_encoder.  
//        1 eccblock = 72 AVIframes (64 data + 8 parity.)
//        *(the first ECCblock in the AVIfile is 'stretched', due to
//          repitition of AVIframe#0.)
//
//  Source Code changes
//  -------------------
//    The addition of ECC-encoder requires changes to dvspoof.  Here
//  is a summary of changes.
//
//    DVSPOOF ID-tag changed from "SPOF" -> "SPOE"
//      This was done to differentiate the AVI-output of this DVSPOOF
//    from that of the original DVSPOOF.  As you may have guessed,
//    the two are mutually incompatible.
//
//    FindFiles() - adjusted the calculation of FilesInfo.FramesForInputFile
//      to include ECC_parity frames.  The 'dir' routine has been changed to
//      supports to use "", to allow longfilenames.
//    ReadData() - moved the checksum calculation to FillDVFrame()
//    FillDVFrame() - checksum calculation (originally from ReadData())
//      This allows ECC_generated data to be properly checksum'd.
//    WriteDV() - broke-off into a 2nd function "CloseAVIfile()"
//      The main() loop needs to call WriteDVFrame() and CloseAVIfile() 
//      at different times.  CloseAVIfile() is called ONCE per ECC_block,
//      to prevent dvspoof from prematurely closing an AVI-file.
//      Minor bugfixes for size of some local array vars
//    CloseAVIfile() - (new function)
//      This code isn't new, just renamed and put into its own function
//      (from the original DVspoof)  In the old DVspoof this code was
//      run after every WriteDV() call.  Now, this code is only called
//      after an ECC_block completion (immediately after 8-parity frames.)
//    main() - default AVIfile size changed.
//      the main processing-loop has changed quite a bit, too.  
//      Due to the ECC-encoder, the length of AVIfiles is now a multiple of
//      the ECC_block length.  If an AVIfile does not end exactly on
//      an ECC_block boundary, it is automatically zero-padded to fill a
//      complete ECC_block.
//     
//   *** compiled with Microsoft Visual C++ 6.0 pro (with ServicePack5)
//      Microsoft doesn't recognize 'long long', so I have replaced these
//      with __int64.  I hope this doesn't cause a problem.
/////////////////////////////////////////////////////////////////////////////

//typedef unsigned long long _bigint;  // for GCC, (MSVC 6 doesn't like)
typedef __int64 _bigint;  // MS Visual C++ 6.0 Pro, silly rabbit

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <windows.h>

#include "ecc.h"

#define ECC_BLKCTR_MAX 0xFFFFFFFFU

ecc_enc_state g_eccEncoderState;
//ecc_dec_state g_eccDecoderState;

eccword *pECCparitybuf; // ECC_parity data buffer (provided by ecc_encoder)
eccword *pECCinputbuf;  // alias to InputBuffer[]
unsigned char *pECCparitybufC; // unsigned char version of pECCparitybuf


char AVIheader[] = { // This is a Type-2 DV .avi header
    0x52, 0x49, 0x46, 0x46, // 'RIFF'
    0x00, 0x00, 0x00, 0x00, // fill with size of file - 8 bytes
    0x41, 0x56, 0x49, 0x20, // 'AVI ' 
      0x4C, 0x49, 0x53, 0x54, // 'LIST'
      0XC0, 0x00, 0x00, 0x00, // Size of LIST (192 bytes)
      0x68, 0x64, 0x72, 0x6C, // 'hdrl' 
        0x61, 0x76, 0x69, 0x68, // 'avih'
        0x38, 0x00, 0x00, 0x00, // Size of avih (56 Bytes)
          0x57, 0x82, 0x00, 0x00, // dwMicroSecPerFrame    : 33367 (Some files are 33366?)
          0X00, 0x00, 0x00, 0x00, // dwMaxBytesPerSec      : 0 (unfilled)      
          0X00, 0x00, 0x00, 0x00, // dwPaddingGranularity  : 0
          0X10, 0x01, 0x00, 0x00, // dwFlags               : ???
          0X00, 0x00, 0x00, 0x00, // 48,49,50,51 dwTotalFrames   : (FILL ME)
          0x00, 0x00, 0x00, 0x00, // dwInitialFrames       : 0
          0x01, 0x00, 0x00, 0x00, // dwStreams             : 1 (Video Stream only)
              0xC0, 0xD4, 0x01, 0x00, // dwSuggestedBufferSize : 120000
                0xD0, 0x02, 0x00, 0x00, // dwWidth               : 720
          0xE0, 0x01, 0x00, 0x00, // dwHeight              : 480
          0x00, 0x00, 0x00, 0x00, // dwReserved            : 0x0
          0x00, 0x00, 0x00, 0x00, // dwReserved            : 0x0
          0x00, 0x00, 0x00, 0x00, // dwReserved            : 0x0
          0x00, 0x00, 0x00, 0x00, // dwReserved            : 0x0
        0x4C, 0x49, 0x53, 0x54, // 'LIST'
        0x74, 0x00, 0x00, 0x00, // Size of LIST ( 116 bytes)
        0x73, 0x74, 0x72, 0x6C, // 'strl'
          0x73, 0x74, 0x72, 0x68, // 100 'strh'
          0x38, 0x00, 0x00, 0x00, //  Size of strh (56 bytes)
                            0x76, 0x69, 0x64, 0x73, // fccType  : 'vids'
            0x64, 0x76, 0x73, 0x64, // fccHandler   : 'dvsd'
            0x00, 0x00, 0x00, 0x00, // dwFlags  : 0x0
            0x00, 0x00,     // wPriority  : 0
            0x00, 0x00,   // wLanguage  : 0
            0x00, 0x00, 0x00, 0x00, // dwInitialFrames  : 0     
            0xE9, 0x03, 0x00, 0x00, // dwScale  : 1001
            0x30, 0x75, 0x00, 0x00, // dwRate : 30000 (dwRate/dwScale = 29.97)
            0x00, 0x00, 0x00, 0x00, // dwStart  : 0
            0x00, 0x00, 0x00, 0x00, // 140,141,142,143 dwLength : (FILL ME # frames)
            0xC0, 0xD4, 0x01, 0x00, // dwSuggestedBufferSize : 120000
            0x10, 0x27, 0x00, 0x00, // dwQuality  : 10000 (???)
            0x00, 0x00, 0x00, 0x00, // dwSampleSize : 0 
            0x00, 0x00, 0x00, 0x00,
            0xD0, 0x02, 0xE0, 0x01, // rcFrame  : 0,0,720,480
          0x73, 0x74, 0x72, 0x66, // 'strf'
          0x28, 0x00, 0x00, 0x00, // Size of strf (40 bytes)
            0x28, 0x00, 0x00, 0x00, // dwSize   : 40
            0xD0, 0x02, 0x00, 0x00, // dwWidth    : 720
            0xE0, 0x01, 0x00, 0x00, // dwHeight   : 480
            0x01, 0x00,   // wPlanes    : 1
            0x18, 0x00,   // 187 wBitcount    : 24
            0x64, 0x76, 0x73, 0x64, // dwCompression  : 'dvsd'
            0xC0, 0xD4, 0x01, 0x00, // dwSizeImage    : 120000
            0x00, 0x00, 0x00, 0x00, // dXPelsPerMeter       : 0x0
            0x00, 0x00, 0x00, 0x00, // 200 dYPelsPerMeter       : 0x0
            0x00, 0x00, 0x00, 0x00, // dwClrUsed            : 0x0
            0x00, 0x00, 0x00, 0x00, // dwClrImportant       : 0x0
      0x4C, 0x49, 0x53, 0x54, // 'LIST' (of DV frames)
      0XC0, 0x00, 0x00, 0x00, // 216,217,218,219 Size of LIST (FILL ME ((120000+8)*framecount)+4)
      0x6D, 0x6F, 0x76, 0x69  // 'movi'
    };  

char AVIidxfooter1[] = { 
  0x69, 0x64, 0x78, 0x31, // 'idx1'
  0x00, 0x00, 0x00, 0x00, // fill with # frames << 4
  };

char AVIidxfooter2[] = { 
  0x30, 0x30, 0x64, 0x63, // '00dc'
  0x10, 0x00, 0x00, 0x00, // some kind of constant value?
  0x00, 0x00, 0x00, 0x00, // fill with frame offset
  0xC0, 0xD4, 0x01, 0x00  // represents size of DV frame (120,000 bytes)
  };

typedef struct { // Information about this frame
  char DVSpoofTag[4]; 
  unsigned int Checksum;  // Checksum for data in frame (3 bytes) 4-6
  _bigint BytesLeft;  // How many bytes left for this input file (5 bytes) 7-11
  unsigned int FileNum;   // Current File # (2 bytes) 12-13
  unsigned int TotalFiles;  // Total number of input files (2 bytes) 14-15
  unsigned int FrameNum;    // Current Frame # (3 bytes) 16-18
  unsigned int TotalFrames; // Total number of frames for this input file (3 bytes) 19-21
  char FieldTag;
  char VAUXTag; 
  char AudioTag;
  char VideoTag;
} FrameInfoType;

typedef struct { // Information about this file and all files
  int FramesPerOutputFile;// Input Parameter
  int RecurseSubdirectories; // Are we looking into subdirectories for files or not
  int Frame0RepeatCount;  // Number of times Frame0 is to be repeated
  int TotalAVIFiles;  // Total estimate number of output .avi files
  int CurrentAVIFile;
  int FramesInAVIFile[400];// Total Sizes of each AVI file (for AVI header), idx starts at 1
  int CurrentAVIFrameCount; // # of frames already written to in this .avi file
  int TotalInputFiles;  // Total Number of input files
  int CurrentInputFile;
  char NameOfInputFile[400][255]; 
  long SizeOfInputFile[400]; // Total Sizes for each Input File (for DVSpoof header), idx starts at 1
  int FramesForInputFile[400]; // Total number of frames in each AVI file (for AVI header), idx starts at 1
} FilesInfo;

////////////////////////////////////////////////////////////////////////////////////////////
//FindFiles generates the list of input files to be encoded and determines filelengths, etc
//in order to populate the .avi file headers as well as the DVspoof headers
//
//FindFiles makes use of a System DIR call (cheezy, I know) to grab the (recursive) list
//of input file names. This is because MingW32 gcc globbing does not seem to be working properly
//for recursive pattern matching.
////////////////////////////////////////////////////////////////////////////////////////////
int FindFiles (FilesInfo * FilesInfo, unsigned char * dircommand) {
  
  int tempframes = 0;
  int i,filesizepos, ecc_k, last_eccblk_marker;
  _bigint tempincr;
  _bigint tempbytes = 0;
  char line[255];
  char filesize[22];
  char namebuffer[255];
  char dircommand2[255];
  FILE * dirfile;
  FILE * fp = NULL;
  //HANDLE hFile;
  //DWORD dFilesizehigh, dFilesizelow;
  int j;
  int startpos,endpos;

  system (dircommand); // System DIR call into a file called SPOOFDIR
  system ("sort <SPOOFDIR >SPOOFDI3"); // Sort dir listing
  remove ("SPOOFDIR");

  ecc_k = 0;
  last_eccblk_marker = 0;

  dirfile = fopen("SPOOFDI3","r");
  while (fgets(line,255,dirfile) != NULL) {
    if (FilesInfo->RecurseSubdirectories) { // Strip drive letter 
      memcpy(namebuffer,line+2,strlen(line)-3);
      namebuffer[strlen(line)-3] = '\0'; // strip newline char

    }
    else {
      memcpy(namebuffer,line,strlen(line)-1);
      namebuffer[strlen(line)-1] = '\0'; // strip newline char
    }
  
    // Extremely cheesy hack to find filelength using a DIR command!
    sprintf(dircommand2, "dir \"%s\" >SPOOFDI2", namebuffer);
    //strcpy (dircommand2, "dir ");
    //strcat (dircommand2, namebuffer);
    //strcat (dircommand2, " >SPOOFDI2");
    // ignore the SPOOFDIR, DVFRAME_NTSC, and log files
    if (     (strcmp(namebuffer+(strlen(namebuffer)-8),"SPOOFDIR") != 0)   
      && (strcmp(namebuffer+(strlen(namebuffer)-8),"SPOOFDI2") != 0) 
      && (strcmp(namebuffer+(strlen(namebuffer)-8),"SPOOFDI3") != 0) 
      && (strcmp(namebuffer+(strlen(namebuffer)-12),"DVFRAME_NTSC") != 0)
      && (strcmp(namebuffer+(strlen(namebuffer)-12),"DVenclog.txt") != 0)
        ) {     
      system (dircommand2);
      fp = fopen("SPOOFDI2","r");
      fgets(line,255,fp);
      fgets(line,255,fp);
      fgets(line,255,fp);
      fgets(line,255,fp);
      fgets(line,255,fp);
      fgets(line,255,fp);
      startpos = 0;
      for (i=0; i<strlen(line);i++) {
        if (line[i] == ',') {
          for (j=i;j>=0;j--) {
            if(line[j] == ' ') {
              startpos = j;
              j = -1;
            }
          }
          for (j=i;j<strlen(line);j++) {
            if (line[j] == ' ') {
              endpos = j;
              j = strlen(line);
            }
          }
          i = strlen(line);
        }
      }
      memcpy (filesize, line+startpos, endpos-startpos);
      filesize[endpos-startpos] = '\0';
      filesizepos = 0;
      for (i=0; i<20;i++) {
        if ((filesize[i] != ',') && (filesize[i] != ' ')) {
          filesize[filesizepos] = filesize[i];
          filesizepos++;
        }
      }
      filesize[filesizepos] = '\0';
      filesizepos = 0;
      tempbytes = 0;
      for (j=strlen(filesize)-1;j>=0;j--) {
        tempincr = (filesize[j]-48)*(pow(10,filesizepos));
        tempbytes = tempbytes + tempincr;
        filesizepos++;
      }
      fclose(fp);
      remove ("SPOOFDI2");
    }

// Windows specific code to get Filelength INCOMPLETE
//    hFile = CreateFile(namebuffer,
//           GENERIC_READ,
//           0,
//           NULL,
//           OPEN_EXISTING,
//           FILE_ATTRIBUTE_NORMAL,
//           NULL);
//    
//    dFilesizelow = GetFileSizeEx (hFile,&dFilesizehigh);



    if (strlen(namebuffer) >227) {
      printf ("(Ignoring file, pathname exceeds 227 chars)\n");
    }
    // ignore the SPOOFDIR, DVFRAME_NTSC, and log files
    else if (     (strcmp(namebuffer+(strlen(namebuffer)-8),"SPOOFDIR") == 0)   
      || (strcmp(namebuffer+(strlen(namebuffer)-8),"SPOOFDI2") == 0) 
      || (strcmp(namebuffer+(strlen(namebuffer)-8),"SPOOFDI3") == 0) 
      || (strcmp(namebuffer+(strlen(namebuffer)-12),"DVFRAME_NTSC") == 0)
      || (strcmp(namebuffer+(strlen(namebuffer)-12),"DVenclog.txt") == 0)
         ) { 
  
      // Ignore File
    }
    else if (tempbytes > 0)  { // ignore directories & 0-byte files

      fp = fopen(namebuffer,"rb");
//      ecc_k = 0;

      if (fp != NULL) {
        FilesInfo->TotalInputFiles++;
        strcpy(FilesInfo->NameOfInputFile[FilesInfo->TotalInputFiles],namebuffer);
        FilesInfo->SizeOfInputFile[FilesInfo->TotalInputFiles] = tempbytes; // True number of bytes

        // Determine # of frames this file will take -> Take into account reduced Frame0 size and number of repeated Frame 0's
        FilesInfo->FramesForInputFile[FilesInfo->TotalInputFiles] = FilesInfo->Frame0RepeatCount;
        
        // Frame#0 (fewer bytes than usual)
        tempbytes -= (102524-(14*76));
        ++ecc_k; // frame #0 is only read *ONCE* by the ECC_decoder!
        if ( ecc_k == ECC_K )
        {
          // insert 8 (ECC_NMK) parity frames for every 64 (ECC_K) data frames
          FilesInfo->FramesForInputFile[FilesInfo->TotalInputFiles]+= ECC_NMK;
          ecc_k = 0;
          last_eccblk_marker = FilesInfo->FramesForInputFile[FilesInfo->TotalInputFiles];
        }

        while (tempbytes > 0) {
          FilesInfo->FramesForInputFile[FilesInfo->TotalInputFiles]++;
          tempbytes -= 102524;
          ++ecc_k;
          if ( ecc_k == ECC_K )
          {
            // insert 8 (ECC_NMK) parity frames for every 64 (ECC_K) data frames
            FilesInfo->FramesForInputFile[FilesInfo->TotalInputFiles]+= ECC_NMK;
            ecc_k = 0;
            last_eccblk_marker = FilesInfo->FramesForInputFile[FilesInfo->TotalInputFiles];
          }
        }
        
        if (FilesInfo->FramesInAVIFile[FilesInfo->TotalAVIFiles] +
            FilesInfo->FramesForInputFile[FilesInfo->TotalInputFiles] >
            FilesInfo->FramesPerOutputFile ) { // More frames than will fit in current AVI file
      
          //tempframes = FilesInfo->FramesForInputFile[FilesInfo->TotalInputFiles] -
          //  (last_eccblk_marker - FilesInfo->FramesInAVIFile[FilesInfo->TotalAVIFiles]); // excess

          //FilesInfo-> FramesInAVIFile[FilesInfo->TotalAVIFiles] = last_eccblk_marker;

          tempframes = FilesInfo->FramesForInputFile[FilesInfo->TotalInputFiles] -
            (FilesInfo->FramesPerOutputFile - FilesInfo->FramesInAVIFile[FilesInfo->TotalAVIFiles]); // excess

          FilesInfo-> FramesInAVIFile[FilesInfo->TotalAVIFiles] = FilesInfo->FramesPerOutputFile;
          FilesInfo->TotalAVIFiles++;

          while (tempframes >= FilesInfo->FramesPerOutputFile) {
            FilesInfo-> FramesInAVIFile[FilesInfo->TotalAVIFiles] = FilesInfo->FramesPerOutputFile;
            FilesInfo->TotalAVIFiles++;
            tempframes -= FilesInfo->FramesPerOutputFile;
          }
          FilesInfo-> FramesInAVIFile[FilesInfo->TotalAVIFiles] = tempframes; // spill over excess to new avi file
        }
        else { // no excess
          FilesInfo-> FramesInAVIFile[FilesInfo->TotalAVIFiles] += FilesInfo->FramesForInputFile[FilesInfo->TotalInputFiles];
        }

      }
      fclose(fp);
    }
  }
  fclose(dirfile);

  // 'round up' the AVI-file's framecount to a multiple of the ECC_block 
  //  length (ECC_N).  We only do this for the *LAST* AVI-file, because we
  //  assume that the intermediate AVI-files are *already* block aligned.
  //  (Hopefully FilesInfo->FramesPerOutputFile = a 'nice round value'.
  //   An explanation is in main().)
  if ( (ecc_k > 0) && (ecc_k < ECC_K) )
  {
    printf("\nfindfiles() : rounding up AVIfile[%d] from ecc_k = %d\n", 
      FilesInfo->TotalInputFiles, ecc_k );
    // insert 8 (ECC_NMK) parity frames for every 64 (ECC_K) data frames
    FilesInfo->FramesForInputFile[FilesInfo->TotalInputFiles]+= (ECC_N - ecc_k);

    // update the #AVIframes in output file #[]
    FilesInfo->FramesInAVIFile[FilesInfo->TotalAVIFiles] += (ECC_N - ecc_k);
    ecc_k=0;
  }

  remove("SPOOFDI3");

  return 1;
}

////////////////////////////////////////////////////////////////////////////////////////////////
//ReadData reads a frames worth (or less) of data from the next input file
//ReadData also inserts the extended data into the first 15 DIF blocks of Frame0 (for example, Filename)
////////////////////////////////////////////////////////////////////////////////////////////////
int ReadData(unsigned char * InputBuffer, FrameInfoType * FrameInfo, FilesInfo * FilesInfo, FILE ** infile) {

  int bytesread,i;
//  unsigned int checksum = 0;

  if (*infile == NULL) { // First Frame of the input file, read 14 DIF blocks less
    if (FilesInfo->CurrentInputFile > FilesInfo->TotalInputFiles) { // no more input files
      return 0;
    }
    else {
      *infile = fopen(FilesInfo->NameOfInputFile[FilesInfo->CurrentInputFile],"rb");
      FrameInfo->FileNum = FilesInfo->CurrentInputFile;
      FrameInfo->TotalFiles = FilesInfo->TotalInputFiles;
      FrameInfo->FrameNum = -1;
      FrameInfo->TotalFrames = FilesInfo->FramesForInputFile[FilesInfo->CurrentInputFile] - 1;

      for (i=0;i<(14*76);i++) { InputBuffer[i] = 0x00; }
      bytesread = fread (InputBuffer+(14*76),1,102524-(14*76), *infile);

      // Store Filename in Video DIF blocks 1-3 (228 bytes)
      memcpy (InputBuffer, FilesInfo->NameOfInputFile[FilesInfo->CurrentInputFile], 228);
      InputBuffer[228] = '\0'; // Force string terminator
    }
  }
  else { // Not the first frame, read full frame of data
    bytesread = fread(InputBuffer, 1, 102524, *infile);
  }

  FrameInfo->FrameNum++;
  FrameInfo->BytesLeft = FilesInfo->SizeOfInputFile[FilesInfo->CurrentInputFile];
  FilesInfo->SizeOfInputFile[FilesInfo->CurrentInputFile] -= bytesread;

//  for (i=0;i<102524;i++) {
//    checksum = checksum + InputBuffer[i];
//  }
//  FrameInfo->Checksum = checksum;

  if (FilesInfo->SizeOfInputFile[FilesInfo->CurrentInputFile] == 0) { // Input File is done
    fclose(*infile);
    *infile = NULL;
    FilesInfo->CurrentInputFile++;
  }
  return 1;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
//FillDVFrame takes the raw input data and formats it into valid DV frames, complete with additional
//DVSpoof-specific info
//////////////////////////////////////////////////////////////////////////////////////////////////
int FillDVFrame (unsigned char * InputBuffer, FrameInfoType * FrameInfo, FilesInfo * FilesInfo, unsigned char * DVFrameBuffer) {
 
  int i,j,difblockoffset;
  int inpointer = 0;
  int inpointercount = 0;
  unsigned char tempbuffer[24];
  unsigned int checksum = 0;

  // calculate checksum
  for (i=0;i<102524;i++)
    checksum = checksum + InputBuffer[i];
  FrameInfo->Checksum = checksum;

  // Field Tag (I think?)
  DVFrameBuffer[12] = FrameInfo->FieldTag;

  // VAUX Headers (need slight change/frame)
  DVFrameBuffer[248] = FrameInfo->VAUXTag;
  DVFrameBuffer[328] = FrameInfo->VAUXTag;
  DVFrameBuffer[408] = FrameInfo->VAUXTag;

  // Audio Headers (need slight change/frame)
  for (i=0;i<10;i++) {
    for (j=0;j<9;j++) {
      DVFrameBuffer[(i*12000)+(j*1280)+488] = FrameInfo->AudioTag;
    }
  }
  
  // First video block = DVSpoof header
  tempbuffer[0] = FrameInfo->DVSpoofTag[0];
  tempbuffer[1] = FrameInfo->DVSpoofTag[1];
  tempbuffer[2] = FrameInfo->DVSpoofTag[2];
  tempbuffer[3] = FrameInfo->DVSpoofTag[3];

  tempbuffer[4] = FrameInfo->Checksum;
  tempbuffer[5] = FrameInfo->Checksum >> 8;
  tempbuffer[6] = FrameInfo->Checksum >> 16;

  tempbuffer[7] = FrameInfo->BytesLeft;
  tempbuffer[8] = FrameInfo->BytesLeft >> 8;
  tempbuffer[9] = FrameInfo->BytesLeft >> 16;
  tempbuffer[10] = FrameInfo->BytesLeft >> 24;
  tempbuffer[11] = FrameInfo->BytesLeft >> 32;

  tempbuffer[12] = FrameInfo->FileNum;
  tempbuffer[13] = FrameInfo->FileNum >> 8;
  tempbuffer[14] = FrameInfo->TotalFiles;
  tempbuffer[15] = FrameInfo->TotalFiles >> 8;
  
  tempbuffer[16] = FrameInfo->FrameNum;
  tempbuffer[17] = FrameInfo->FrameNum >> 8;
  tempbuffer[18] = FrameInfo->FrameNum >> 16;
  tempbuffer[19] = FrameInfo->TotalFrames - FilesInfo->Frame0RepeatCount + 1;
  tempbuffer[20] = (FrameInfo->TotalFrames - FilesInfo->Frame0RepeatCount + 1) >> 8;
  tempbuffer[21] = (FrameInfo->TotalFrames - FilesInfo->Frame0RepeatCount + 1) >> 16;

  tempbuffer[22] = 0xFF;
  tempbuffer[23] = 0xFF;

  DVFrameBuffer[568] = FrameInfo->VideoTag;
  // Copy DVSpoof header 3 times into DIF block (cheesy error correction)
  memcpy (DVFrameBuffer+572, tempbuffer,24);
  memcpy (DVFrameBuffer+596, tempbuffer,24);
  memcpy (DVFrameBuffer+620, tempbuffer,24);

  // Fill Actual Data
  for (i=0; i<10; i++) {
    for (j=0;j<150;j++) {
      difblockoffset = (i*12000)+(j*80);
      // Fill every DIF block from 8 onwards except the Audio DIF blocks every 16, and
      // Also fill DIF block 7 from the 2nd DIF sequence onwards 
      // (DVSpoof header is in the first one therefore 1349 Video DIF blocks used)
      if (  ((j>7) && (((j-6) % 16) != 0)) || ( (j==7) && (i!=0) ) )  {
        DVFrameBuffer[8+difblockoffset] = FrameInfo->VideoTag;
        memcpy(DVFrameBuffer+8+difblockoffset+4,InputBuffer+inpointer,76);

        inpointer +=76;
        inpointercount++;
      }
    }
  }

  if (FrameInfo->FieldTag == 0x68) { // Flip field bit every frame
    FrameInfo->FieldTag = 0x78;
  }
  else { 
    FrameInfo->FieldTag = 0x68;
  }
  FrameInfo->VAUXTag++;
  FrameInfo->AudioTag++;
  FrameInfo->VideoTag++;
  if (FrameInfo->VAUXTag > 0x5b) {
    FrameInfo->VAUXTag = 0x50;
    FrameInfo->AudioTag = 0x70;
    FrameInfo->VideoTag = 0x90;
  }

  return 1;
}



void 
CloseAVIfile( FilesInfo * FilesInfo, FILE ** outfile)
{
  int i;

    if (FilesInfo->CurrentAVIFrameCount >= FilesInfo->FramesInAVIFile[FilesInfo->CurrentAVIFile] ) { // Close .avi file
      printf("\nCloseAVIfile() : closing AVIfile[%d] at frame#%d!", 
        FilesInfo->CurrentAVIFile, FilesInfo->CurrentAVIFrameCount );
      
      // check 
      if (FilesInfo->CurrentAVIFrameCount > FilesInfo->FramesInAVIFile[FilesInfo->CurrentAVIFile] )
        printf( "\nCloseAVIfile() : UH OH, AVIfile[%d] should only have %d frames!",
          FilesInfo->CurrentAVIFile, 
          FilesInfo->FramesInAVIFile[FilesInfo->CurrentAVIFile]);

      // Write out .avi Footer first (idx)
      AVIidxfooter1[4] = FilesInfo->CurrentAVIFrameCount << 4;
      AVIidxfooter1[5] = FilesInfo->CurrentAVIFrameCount >> 4;
      AVIidxfooter1[6] = FilesInfo->CurrentAVIFrameCount >> 12;
      AVIidxfooter1[7] = FilesInfo->CurrentAVIFrameCount >> 20;

      fwrite (AVIidxfooter1, sizeof(AVIidxfooter1),1,*outfile);
      for (i=0; i<FilesInfo->CurrentAVIFrameCount; i++) {
        AVIidxfooter2[8] = ((120008*i)+4);
        AVIidxfooter2[9] = ((120008*i)+4) >> 8;
        AVIidxfooter2[10] = ((120008*i)+4) >> 16;
        AVIidxfooter2[11] = ((120008*i)+4) >> 24;
        fwrite (AVIidxfooter2, sizeof(AVIidxfooter2),1,*outfile);
      }
    
      fclose (*outfile);
      *outfile = NULL;
      FilesInfo->CurrentAVIFile++;
      FilesInfo->CurrentAVIFrameCount = 0;
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//WriteDV manages the output .avi files and writes DV frames to them (including repeating frames as 
//necessary)
///////////////////////////////////////////////////////////////////////////////////////////////////////
int WriteDV(unsigned char * DVFrameBuffer, FrameInfoType * FrameInfo, FilesInfo * FilesInfo, FILE ** outfile) {
static int lctr = 0;
  char filename1[] = { 'D','V','\0' };
  char filename2[] = { '.','a','v','i','\0' };
  char filenamebuffer[255]; // 12 is too small
  char filenumbuffer[8];    // 6 is too small
  //char filenamebuffer[12];
  //char filenumbuffer[6];
  int framecount,j,repeatcount;
  
  // Establish # of times to repeat this frame
  if (FrameInfo->FrameNum == 0) {
    repeatcount = FilesInfo->Frame0RepeatCount;
  }
  else {
    repeatcount = 1;
  }
++lctr;
  for (j=0;j<repeatcount;j++) {

    if (*outfile == NULL) { // open a new file
      strcpy (filenamebuffer,filename1);
      sprintf (filenumbuffer,"%06d",FilesInfo->CurrentAVIFile);
      strcat (filenamebuffer,filenumbuffer);
      strcat (filenamebuffer,filename2);

      *outfile = fopen (filenamebuffer, "wb");
      printf ("(Outputting to %s)",filenamebuffer);

      FilesInfo->CurrentAVIFrameCount = 0;

      //Create .avi file header
      framecount = FilesInfo->FramesInAVIFile[FilesInfo->CurrentAVIFile];

      //Total length of .avi file in bytes
      AVIheader[4] = (212 + ((((120000+8)*framecount)+4) + (8 + (16*framecount))));
      AVIheader[5] = (212 + ((((120000+8)*framecount)+4) + (8 + (16*framecount)))) >> 8;
      AVIheader[6] = (212 + ((((120000+8)*framecount)+4) + (8 + (16*framecount)))) >> 16;
      AVIheader[7] = (212 + ((((120000+8)*framecount)+4) + (8 + (16*framecount)))) >> 24;

      //Total # of frames in .avi file
      AVIheader[48] = framecount;
      AVIheader[49] = framecount >> 8;
      AVIheader[50] = framecount >> 16;
      AVIheader[51] = framecount >> 24;
      AVIheader[140] = framecount;
      AVIheader[141] = framecount >> 8;
      AVIheader[142] = framecount >> 16;
      AVIheader[143] = framecount >> 24;

      //Total # of bytes in main LIST construct
      AVIheader[216] = (((120000+8)*framecount)+4);
      AVIheader[217] = (((120000+8)*framecount)+4) >> 8;
      AVIheader[218] = (((120000+8)*framecount)+4) >> 16;
      AVIheader[219] = (((120000+8)*framecount)+4) >> 24;
      fwrite (AVIheader, sizeof(AVIheader), 1, *outfile);
    }

    // Write out DV Frame 
    fwrite (DVFrameBuffer, 120008, 1, *outfile);
    FilesInfo->CurrentAVIFrameCount++;
    CloseAVIfile( FilesInfo, outfile );

/* this code has been moved to a new function CloseAVIfile()
    if (FilesInfo->CurrentAVIFrameCount == FilesInfo->FramesInAVIFile[FilesInfo->CurrentAVIFile] ) { // Close .avi file

      // Write out .avi Footer first (idx)
      AVIidxfooter1[4] = FilesInfo->CurrentAVIFrameCount << 4;
      AVIidxfooter1[5] = FilesInfo->CurrentAVIFrameCount >> 4;
      AVIidxfooter1[6] = FilesInfo->CurrentAVIFrameCount >> 12;
      AVIidxfooter1[7] = FilesInfo->CurrentAVIFrameCount >> 20;

      fwrite (AVIidxfooter1, sizeof(AVIidxfooter1),1,*outfile);
      for (i=0; i<FilesInfo->CurrentAVIFrameCount; i++) {
        AVIidxfooter2[8] = ((120008*i)+4);
        AVIidxfooter2[9] = ((120008*i)+4) >> 8;
        AVIidxfooter2[10] = ((120008*i)+4) >> 16;
        AVIidxfooter2[11] = ((120008*i)+4) >> 24;
        fwrite (AVIidxfooter2, sizeof(AVIidxfooter2),1,*outfile);
      }
    
      fclose (*outfile);
      *outfile = NULL;
      FilesInfo->CurrentAVIFile++;
      printf("\nClosed file!");
    }
*/
  }
  return 1; 
}



// flush_eccblock() is called when the AVIfile reaches EOF.  this function
// takes an 'incomplete' ecc_block, and fills it with 00's, so that the
// very last part of the AVIfile forms a complete ECC-block.
void
flush_eccblock( unsigned char * DVFrameBuffer,  unsigned char *InputBuffer,
            FrameInfoType * FrameInfo, FilesInfo * FilesInfo, 
            FILE ** outfile, const int k)
{
  int ecc_pctr = k;
  int error, rc;
   
  printf("\nflush_eccblock( ecc_k=%d ) : starting (CAVIFcnt=%d)", k,
    FilesInfo->CurrentAVIFrameCount );

  if ( (ecc_pctr > 0) && (ecc_pctr < ECC_K) )
  {
    printf("\nflush_eccblock() : padding file to ECC_block boundary! (ecc_pctr=%d)", ecc_pctr);
    // clear the memory buffer
    memset( (void *) InputBuffer, 0, sizeof(eccword)*ECC_Z );
    // mark these frames as 'zero-padding' 
    FrameInfo->BytesLeft = 0;

    // fill the remaining ECC_block with 'zero'd frames 
    for ( ; ecc_pctr < ECC_K; ++ecc_pctr )
    {
      FrameInfo->FrameNum++;

      // Populate actual DV Frame with input data
      error = FillDVFrame( InputBuffer, FrameInfo, FilesInfo, DVFrameBuffer);
      rc = ecc_encoder_enc1( &g_eccEncoderState, pECCinputbuf );
      error = WriteDV(DVFrameBuffer, FrameInfo, FilesInfo, outfile);
    } // end for ( ; ecc_pctr < ECC_K ...

    // get the parity frames, and write them to the AVI-file
    for ( ecc_pctr = 0; ecc_pctr < ECC_NMK; ++ecc_pctr )
    {
      // retrieve 1 parity frame
      rc = ecc_encoder_getp1( &g_eccEncoderState, &pECCparitybuf );
      pECCparitybufC = (unsigned char *) pECCparitybuf;
      FrameInfo->FrameNum++;

      // Populate actual DV Frame with input data
      error = FillDVFrame( pECCparitybufC, FrameInfo, FilesInfo, DVFrameBuffer);

      printf ("\nEncoding parity %d",FrameInfo->FrameNum );

      // update the FrameInfo header
      // write 1 parity frame
      error = WriteDV(DVFrameBuffer, FrameInfo, FilesInfo, outfile);
    } // end for ( ecc_pctr = 0 ...

    // update counters
    ecc_pctr = 0;
    //if ( ecc_blkctr < ECC_BLKCTR_MAX )
//      ++ecc_blkctr; // increment ECCblock counter

    // close the AVIfile (add the footer)
      //CloseAVIfile( FilesInfo, outfile); // only closes AVI-file if necessary
  } // end if ( ecc_pctr ... )
}


// NOTE: THIS TURNS OFF GLOBBING SUPPORT UNDER MINGW32 COMPILER. GLOBBING MUST BE TURNED OFF!
int _CRT_glob = 0; 

////////////////////////////////////////////////////////////////////////////////////////////
//Main
////////////////////////////////////////////////////////////////////////////////////////////
int main(int argc, char * argv[]) {

  FrameInfoType FrameInfo;
  FilesInfo FilesInfo;

  unsigned char DVFrameBuffer[120008];
  unsigned char InputBuffer[102524];

  int error,i,j;
  int ecc_pctr;// ECC bit-plane counter, 0..63
  unsigned int ecc_blkctr; // ECC block-counter, 0..XXX
  int rc; // function return code

  FILE * DVFramefile;
  FILE * logfile;
  FILE * infile = NULL;
  FILE * outfile = NULL;
  
  char dircommand[255];

  // initialize the ECC-encoder and decoder
    // initialize the ECC_decoder
  g_eccEncoderState.pMessagePlane = 0; // initializing for first time
  for ( i = 0; i < ECC_NMK; ++i )
    g_eccEncoderState.pParityPlanes[i] = 0; // initializing for first time
  ecc_encoder_init( &g_eccEncoderState );

  pECCinputbuf = (eccword *) InputBuffer; // setup alias to Framebuffer

  if (argc == 1) { 
    printf ("DVenc v2.0 BETA - By Andy Kromkamp\n");
    printf ("  (Hamming-code ECC by Royce Liao\n");
    printf ("Usage: DVenc [-recurse] [-repeat0 # frames] [-avisize # frames] <input files>\n\n");
    printf ("  -recurse: Search for input files recursively through directories. Note that\n");
    printf ("            entire directory path from the root of the hard drive (excluding\n");
    printf ("            the drive letter is stored.\n");
    printf ("  -repeat0: Specifies number of times to repeat Frame 0 (to guarantee it is\n");
    printf ("            retreived without errors). Defaults to 30, and cannot be set lower\n");
    printf ("		         than 10.  (Do not set higher than 80)\n");
    printf ("  -avisize: Specifies number of frames per output .avi file. Defaults to 15870,\n");
    printf ("            which is approximately 2 GB. ***DO NOT CHANGE***\n\n");
    printf ("Example: DVenc *.mpg\n");
    printf ("Example: DVenc file1.mpg file2.mpg\n");
    printf ("Example: DVenc -recurse *.mpg\n");
    printf ("Example: DVenc -recurse *.ts.*\n");
    printf ("Example: DVenc -repeat0 15 hipix.ts.*\n");
    printf ("Example: DVenc -avisize 500 hipix.ts.*\n\n");
    printf ("NOTE: PAL Camcorders are not supported at this time.\n");
    printf ("NOTE: Multiple input files on the command line (ie. DVenc file*.mpg data*.mpg)\n");
    printf ("      are not supported under Win98 (Although wildcards are)\n");
    return (-1); 
  } 

  // DVSpoofTag changed from "SPOF" to "SPOE"
  FrameInfo.DVSpoofTag[0] = 'S';
  FrameInfo.DVSpoofTag[1] = 'P';
  FrameInfo.DVSpoofTag[2] = 'O';
  FrameInfo.DVSpoofTag[3] = 'E';
  FrameInfo.FieldTag = 0x68;
  FrameInfo.VAUXTag = 0x56;
  FrameInfo.AudioTag = 0x76;
  FrameInfo.VideoTag = 0x96;

  //FilesInfo.FramesPerOutputFile = 16000;
  FilesInfo.RecurseSubdirectories = 0;
  FilesInfo.Frame0RepeatCount = 30;
  FilesInfo.TotalAVIFiles = 1;
  FilesInfo.FramesInAVIFile[1] = 0;
  FilesInfo.TotalInputFiles = 0;
  // 15840 = 220*64 data-frames + 220*8 parity
  //FilesInfo.FramesPerOutputFile = 144 + (FilesInfo.Frame0RepeatCount - 1);
  FilesInfo.FramesPerOutputFile = 15840 + (FilesInfo.Frame0RepeatCount - 1);

  for (i=1; i< argc; i++) { 
    if (!strcmp(argv[i],"-recurse")) {
      FilesInfo.RecurseSubdirectories = 1;
    }
    else if (!strcmp(argv[i],"-repeat0")) {
      FilesInfo.Frame0RepeatCount = atoi(argv[i+1]);
      i++;
    }
    else if (!strcmp(argv[i],"-avisize")) {
      FilesInfo.FramesPerOutputFile = atoi(argv[i+1]);
      i++;
    }
    else { // Must be filenames
      strcpy (dircommand,"dir /B /ON ");
      if (FilesInfo.RecurseSubdirectories == 1) {
        strcat (dircommand, "/S ");
      }
      for (j=i;j<argc;j++) {
        strcat (dircommand, argv[j]);
        strcat (dircommand, " ");
      }
      strcat (dircommand, " >SPOOFDIR");
      i = argc;
    }
  }
  
  // Populate DVFrame with stuff relevant to the DV spec
  DVFramefile = fopen ("DVFRAME_NTSC","rb");
  if (DVFramefile == NULL) {
    printf ("Cannot find DVFRAME_NTSC File!\n");
    return (-1);
  }
  fread(DVFrameBuffer,1,120008,DVFramefile);
  error = fclose(DVFramefile);

  logfile = fopen ("DVenclog.txt","w");
  if (logfile == NULL) {
    printf ("Could not open Logfile!\n");
    return (-1);
  }

  // Parse Directory to collect information about all input files (TODO: RECURSIVE, WILDCARD)
  error = FindFiles(&FilesInfo,dircommand);

  if (FilesInfo.TotalInputFiles == 0) {
    printf ("DVenc v2.0 BETA - By Andy Kromkamp\n\n");
    printf ("Usage: DVenc [-recurse] [-repeat0 # frames] [-avisize # frames] <input files>\n\n");
    printf ("  -recurse: Search for input files recursively through directories. Note that\n");
    printf ("            entire directory path from the root of the hard drive (excluding\n");
    printf ("            the drive letter is stored.\n");
    printf ("  -repeat0: Specifies number of times to repeat Frame 0 (to guarantee it is\n");
    printf ("            retreived without errors). Defaults to 30, and cannot be set lower\n");
    printf ("		   than 10.\n");
    //printf ("  -avisize: Specifies number of frames per output .avi file. Defaults to 16000,\n");
    printf ("  -avisize: Specifies number of frames per output .avi file. Defaults to 15984,\n");
    printf ("            which is approximately 2 GB. (Must be divisible by 72!)\n\n");
    printf ("Example: DVenc *.mpg\n");
    printf ("Example: DVenc file1.mpg file2.mpg\n");
    printf ("Example: DVenc -recurse *.mpg\n");
    printf ("Example: DVenc -recurse *.ts.*\n");
    printf ("Example: DVenc -repeat0 15 hipix.ts.*\n");
    printf ("Example: DVenc -avisize 500 hipix.ts.*\n\n");
    printf ("NOTE: PAL Camcorders are not supported at this time.\n");
    printf ("NOTE: Multiple <input files> on the command line (ie. DVenc file*.mpg data*.mpg)\n");
    printf ("      are not supported under Win98 (Although wildcards are)\n");

    return (-1); 
  }

  for (i=1;i<=FilesInfo.TotalInputFiles;i++) {
    printf ("Input File[%d] %s: %u Bytes -> %d Frames\n",i,
      FilesInfo.NameOfInputFile[i],FilesInfo.SizeOfInputFile[i],
      FilesInfo.FramesForInputFile[i]);
    fprintf (logfile, "Input File[%d] %s: %u Bytes -> %d Frames\n",i,
      FilesInfo.NameOfInputFile[i],FilesInfo.SizeOfInputFile[i],
      FilesInfo.FramesForInputFile[i]);
  }

  for (i=1;i<=FilesInfo.TotalAVIFiles;i++) {
    printf ("AVI File %d: %d Frames\n",i,FilesInfo.FramesInAVIFile[i]);
    fprintf (logfile,"AVI File %d: %d Frames\n",i,FilesInfo.FramesInAVIFile[i]);
  }

  printf ("(Frame 0 Repeated %d times)\n",FilesInfo.Frame0RepeatCount);
  fprintf (logfile,"(Frame 0 Repeated %d times)\n",FilesInfo.Frame0RepeatCount);

  FilesInfo.CurrentAVIFile = 1;
  FilesInfo.CurrentInputFile = 1;
  FilesInfo.CurrentAVIFrameCount = 0;
 
  ecc_pctr = 0; // reset counter
  ecc_blkctr = 0; // reset Block counter

  // Call ReadData to grab the next 102524 bytes of input data into the TempBuffer
  while (ReadData(InputBuffer,&FrameInfo,&FilesInfo,&infile)) {

    if (FrameInfo.FrameNum == 0) {
      printf ("\n(@ f#%d) Encoding %s",FilesInfo.CurrentAVIFrameCount, 
        FilesInfo.NameOfInputFile[FrameInfo.FileNum]);
      fprintf ( logfile, "\n(@ f#%d) Encoding %s",FilesInfo.CurrentAVIFrameCount, 
        FilesInfo.NameOfInputFile[FrameInfo.FileNum]);
    }
    else {
      printf (".");
      fprintf (logfile,".");
    }
    
    // Populate actual DV Frame with input data
    error = FillDVFrame(InputBuffer, &FrameInfo, &FilesInfo, DVFrameBuffer);
    rc = ecc_encoder_enc1( &g_eccEncoderState, pECCinputbuf );

    // Write out frame into .avi file
    error = WriteDV(DVFrameBuffer, &FrameInfo, &FilesInfo, &outfile);
    //CloseAVIfile( &FilesInfo, &outfile); // only closes AVI-file if necessary
    ++ecc_pctr;

    // Fill 1 ECC-block yet?
    if ( ecc_pctr == ECC_K )
    {
      // get the parity frames, and write them to the AVI-file
      for ( ecc_pctr = 0; ecc_pctr < ECC_NMK; ++ecc_pctr )
      {
        // retrieve 1 parity frame
        rc = ecc_encoder_getp1( &g_eccEncoderState, &pECCparitybuf );
        pECCparitybufC = (unsigned char *) pECCparitybuf;
        FrameInfo.FrameNum++;

        // Populate actual DV Frame with input data
        error = FillDVFrame( pECCparitybufC, &FrameInfo, &FilesInfo, DVFrameBuffer);

        //printf ("\nEncoding parity %d",FrameInfo.FrameNum );

        // update the FrameInfo header
        // write 1 parity frame
        error = WriteDV(DVFrameBuffer, &FrameInfo, &FilesInfo, &outfile);
        //CloseAVIfile( &FilesInfo, &outfile); // only closes AVI-file if necessary

      } // end for ( ecc_pctr = 0 ...

      // update counters
      ecc_pctr = 0;       
      if ( ecc_blkctr < ECC_BLKCTR_MAX )
        ++ecc_blkctr; // increment ECCblock counter

      // close the AVIfile (add the footer)
      //CloseAVIfile( &FilesInfo, &outfile); // only closes AVI-file if necessary

    } // end if ( ecc_pctr >= ECC_K )

    if ( ecc_pctr > ECC_K )
    {
      printf( "main() : CATASTROPHIC ERROR, ecc_pctr(%d) > %d", 
        ecc_pctr, ECC_K );
      fprintf( logfile, "main() : CATASTROPHIC ERROR, ecc_pctr(%d) > %d", 
        ecc_pctr, ECC_K );
    }

  } // while( ...
  
  // Closeout the last ECCblock, if necessary
  if ( (ecc_pctr > 0) && (ecc_pctr < ECC_K) )
  {
    flush_eccblock( DVFrameBuffer,  InputBuffer, &FrameInfo, &FilesInfo, &outfile, 
      ecc_pctr);
    ecc_pctr = 0;
    ++ecc_blkctr;
  }

  printf ("\n\nDone!\n\n");
  printf ("\nECC_encoder wrote %u blocks! (1block = %d AVIframes)", 
            ecc_blkctr, ECC_N );
  fprintf (logfile,"\n\nDone!\n\n");
  fprintf (logfile,"\nECC_encoder wrote %u blocks! (1block = %d AVIframes)", 
            ecc_blkctr, ECC_N );

  fclose(logfile);
  ecc_encoder_shutdown( &g_eccEncoderState );
  return 1;
}