#include<stdlib.h>
#include<stdio.h>
#include<memory.h>
#include<malloc.h>

#include "ecc.h"


//////////////////////////////////////////////////////////////////////////////
//
//    This source code assumes the reader has some knowledge of Hamming
// parity codes.  A good reference book is "Error Control Coding : 
// Fundamentals and Applications" by Shu Lin & Daniel J. Costello.
//
//    This source code is released under the GNU Public License.  For
// details, see http://www.gnu.org
//    If you have questions about *THIS* source-code (and not ECC concepts
// in general -- I'm not expert!), you may email the author at 
// liaor@iname.com
//
// ECC_bit definitions
// -------------------
//     The ECC-encoder operates on groups of individual bits.  The input
//   unit is a packet of 64-databits.  The encoder calculates 8 parity bits
//   for these 64-databits.  The ECC-encoder output unit is 1 'codeword' 
//   (72-codebits = 64 data + 8 parity.) 
//
//             bit organization of 1 codeword
//
//   |<------------------N-------------------->|    N = #codeword (total) bits
//
//   |<------ K ------>| |<---- (N - K) ------>|    K = #data bits (message)
//bit#
//   0 1 2 3 4 ... 62 63 64 65 66 67 68 69 70 71   
//   |<--- user data ->| |<- generated parity->|   N-K= #parity bits
//
// ECC_byte/ECC_word packing
// --------------------------
//
//     It is more efficient for computers to processes words (16, 32, 64 bits)
//   as opposed to individual bits.  So, a practical software ECC_encoder
//   will work in units of bytes/words.  There are two ways to pack data :
//     (1) serially - The ECC_encoder input processes one 1-message at 
//                    a time.  The ECC_encoder output is 1-codeword.
//                    (For example, let's say the ECC_encoder accepts bytes.
//                     Databits are shifted in one at a time, until a 
//                    container byte/word is full.  For a message-length of
//                    64-bit, 8 bytes are needed.  Therefore, each 
//                    use of the ECC_encoder consumes 8-bytes.  And each
//                    use of the ECC_encoder produces 9-bytes, which is 
//                    1 codeword.)
//
//     (2) parallel - The ECC_encoder input processes many messages at a time.
//                    Databits from different codewords are packed together.
//                    Each container byte/word contains just 1 bit from
//                    a given codeword. (For example, let's say the 
//                    ECC_encoder accepts bytes.  Each use of the ECC_encoder
//                    accepts 64-bytes, which consist of 8 independent 
//                    messages.  Each use of the ECC_encoder produces
//                    72-bytes, which is 8 codewords.)
// 
//    This ECC package uses approach (2).  See below
//
//          BYTE-packing of codewords : 8-codewords across 72 bytes
//
//    |<---------------------N----------------------->| N = #codeword (total)
//
//    |<------ K ------------>| |<---- (N - K) ------>| K = #data (message)
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+
//    |a0 | |a1 | ... a62| a63| a64| a65| ... a70| a71| codeword_0 ---
//    |b0 | |b1 | ... b62| b63| b64| b65| ... b70| b71| codeword_1  ^
//    |c0 | |c1 | ... c62| c63| c64| c65| ... c70| c71| codeword_2  |
//    |d0 | |d1 | ... d62| d63| d64| d65| ... d70| d71| codeword_3  |
//    |e0 | |e1 | ... e62| e63| e64| e65| ... e70| e71| codeword_4  |
//    |f0 | |f1 | ... f62| f63| f64| f65| ... f70| f71| codeword_5  |
//    |g0 | |g1 | ... g62| g63| g64| g65| ... g70| g71| codeword_6  V
//    |h0 | |h1 | ... h62| h63| h64| h65| ... h70| h71| codeword_7 ---
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+
//    byte0 byte1      62   63   64   65       70   71
//    |<----- user data ----->| |<- generated parity->| N-K= #parity 
//
//     codeword_0 = bits a0..a71      codeword_4 = bits e0..e71
//     codeword_1 = bits b0..b71      codeword_5 = bits f0..f71
//     codeword_2 = bits c0..c71      codeword_6 = bits g0..g71
//     codeword_3 = bits d0..d71      codeword_7 = bits h0..h71
//
//      BYTE[0] = {h0 g0 f0 e0 d0 c0 b0 a0}
//      BYTE[1] = {h1 g1 f1 e1 d1 c1 b1 a1}
//      BYTE[2] = {h2 g2 f2 e2 d2 c2 b2 a2}
//       ...                  ...
//
//      The ECC_encoder in this package uses 32-bit doublewords.  
//     This means 32-codewords are packed across 72 doublewords.  The
//     user-defined type 'eccword' = unsigned int.
//     (A hypothetical MMX-implementation could pack 64-codewords across 
//     72 quadwords.  And an SSE2-implementation could pack 128-codewords 
//     across 144 quadwords.)
//
// ECC_byte/ECC_word plane packing
// -------------------------------
//     Plane-packing takes the previous data-packing scheme one step 
//   further.  Once again, the ECC_encoder operates on many independent
//   messages per call.  In this example, the databits are packed 
//   across 1-byte (parallel), and 3-bytes form 1 'plane.'  This means
//   the data-structure packs a total of 8 * 3 = 24 messages, across
//   3 * 64 = 192 bytes.  The ECC_encoder generates an output of
//   24 codewords, packed across 3 * 72 = 216 bytes.
//
//    |<---------------------N----------------------->| N = #codeword (total)
//
//    |<------ K ------------>| |<---- (N - K) ------>| K = #data (message)
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+
//    |a0 | |a1 | ... a62| a63| a64| a65| ... a70| a71| codeword_0 ---
//                                                          ...     |
//    |h0 | |h1 | ... h62| h63| h64| h65| ... h70| h71| codeword_7  |
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+             |
//    byte0 byte3                                                   |
//                                                                  |
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+             |
//    |a0 | |a1 | ... a62| a63| a64| a65| ... a70| a71| codeword_8   
//                                                          ...     Z
//    |h0 | |h1 | ... h62| h63| h64| h65| ... h70| h71| codeword_15 
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+             |
//    byte1 byte4                                                   | 
//                                                                  |
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+             |
//    |a0 | |a1 | ... a62| a63| a64| a65| ... a70| a71| codeword_16 |
//                                                          ...     |
//    |h0 | |h1 | ... h62| h63| h64| h65| ... h70| h71| codeword_24 |
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+             |
//    byte2 byte5                                                  ---
//    |<----- user data ----->| |<- generated parity->| N-K= #parity 
//                                                      Z = 3 ('plane depth')
// 
//    Plane-packing is useful for distributing message-bits over a large
//  bitfield.  Assuming the data is serialized as shown (i.e. transmitted in 
//  the described byte-order), an ECC-codeword can tolerate a single 
//  burst error up to Z-bytes in length, and still be correctly decoded.
//
// Error correction performance of ECC_byte/ECC_word plane packing
// ---------------------------------------------------------------
//    How well can the ECC-system correct errors?  To answer this question,
//  let me start by describing the base error-correction algorithm.  
//    This ECC-package implements a SECDED-72,64 Hamming-code.  
//  The Hamming-encoder accepts 64 data bits,
//  and appends 8 parity-bits, outputnig a codeword of 72-bits.  The
//  chosen algorithm has these properties : 1) With 100% accuracy, the
//  Hamming-decoder is guaranteed to correct any single-bit error 
//  within a codeword, and guaranteed to detect any two single-bit errors 
//  within a codeword.
//    I implemented a conservative decoder : it will perform a correction 
//  operation (i.e. change the received codeword) if and only if it detects 
//  a single-bit error.  If the decoder detects > 1-bit error, no action 
//  is taken.  Mathematically, there are some error-patterns (with >1 bit)
//  which can be detected with >50% accuracy, but as stated earlier,
//  the decoder takes action only on single-bit errors.
//    However, the encoder-system does not operate in serial-mode.  It
//  operates on a large word-packed plane data structure.  While data-packing
//  does NOT change the correction-power of the Hamming-code, it can take
//  advantage of some statistical observations.  If burst errors (of length X)
//  are more common that random-errors (of length X), and the error-length
//  is shorter than the plane-depth (Z), then a parallel data-packing 
//  structure will yield better error-performance.  On the other hand, if 
//  errors are distributed evenly across the entire codeword data-structure,
//  and single-bit errors are the most common error, then a serial 
//  data-packing structure will yield better performance.  
//    Therefore, in a practical setting, the correction-performance of 
//  this package's ECC-receiver depends on the system's noise 
//  characteristics.  The system performs best with isolated burst-errors
//  of length < Z, and worse with evenly distributed random bit errors.
//    In the example below, the ECC-decoder operates on 3*72=216 byte
//  codewords.  The decoder can correct any single burst error of 
//  length 3 bytes or less (in the absence of other errors.), or any single 
//  bit-error within a single codeword (codeword_0, codeword_1, etc.)  
//
//    |<---------------------N----------------------->| N = #codeword (total)
//    |<------ K ------------>| |<---- (N - K) ------>| K = #data (message)
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+
//    |a0 | |a1 | ... a62| a63| a64| a65| ... a70| a71| codeword_0 ---
//                                                          ...     |
//    |h0 | |h1 | ... h62| h63| h64| h65| ... h70| h71| codeword_7  |
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+             |
//    byte0 byte3                                                   |
//                                                                  |
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+             |
//    |a0 | |a1 | ... a62| a63| a64| a65| ... a70| a71| codeword_8   
//                                                          ...     Z
//    |h0 | |h1 | ... h62| h63| h64| h65| ... h70| h71| codeword_15 
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+             |
//    byte1 byte4                                                   | 
//                                                                  |
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+             |
//    |a0 | |a1 | ... a62| a63| a64| a65| ... a70| a71| codeword_16 |
//                                                          ...     |
//    |h0 | |h1 | ... h62| h63| h64| h65| ... h70| h71| codeword_24 |
//    +---+ +---+     +--+ +--+ +--+ +--+     +--+ +--+             |
//    byte2 byte5                                                  ---
//    |<----- user data ----->| |<- generated parity->| N-K= #parity 
//                                                      Z = 3 ('plane depth')
//
//////////////////////////////////////////////////////////////////////////////


/*
   from "Error Control Coding : Fundamentals and Applications". Shu Lin,
         Daniel J. Costello.  Prentice Hall PTR 1983.

   (n k) = (72 64)  SECDED Hamming code 

   This code can correct any single-bit error and detect any 2 single-bit
   errors, in a codeword of 72 bits.  PC memory systems use this type of
   error correction (though perhaps not the same codematrix listed below.)

static const unsigned char ecc_matrix_generate[72*8] =
{
 1,1,1,1,1,1,1,1, 0,0,0,0,0,1,1,1, 0,0,1,1,1,0,0,0, 1,1,0,0,1,0,0,0, 0,0,0,0,1,0,0,0, 0,0,0,0,1,0,0,1, 1,0,0,1,0,0,1,0, 0,1,1,0,0,1,0,0, 1,0,0,0,0,0,0,0,
 0,1,1,0,0,1,0,0, 1,1,1,1,1,1,1,1, 0,0,0,0,0,1,1,1, 0,0,1,1,1,0,0,0, 1,1,0,0,1,0,0,0, 0,0,0,0,1,0,0,0, 0,0,0,0,1,0,0,1, 1,0,0,1,0,0,1,0, 0,1,0,0,0,0,0,0,
 1,0,0,1,0,0,1,0, 0,1,1,0,0,1,0,0, 1,1,1,1,1,1,1,1, 0,0,0,0,0,1,1,1, 0,0,1,1,1,0,0,0, 1,1,0,0,1,0,0,0, 0,0,0,0,1,0,0,0, 0,0,0,0,1,0,0,1, 0,0,1,0,0,0,0,0,
 0,0,0,0,1,0,0,1, 1,0,0,1,0,0,1,0, 0,1,1,0,0,1,0,0, 1,1,1,1,1,1,1,1, 0,0,0,0,0,1,1,1, 0,0,1,1,1,0,0,0, 1,1,0,0,1,0,0,0, 0,0,0,0,1,0,0,0, 0,0,0,1,0,0,0,0,
 0,0,0,0,1,0,0,0, 0,0,0,0,1,0,0,1, 1,0,0,1,0,0,1,0, 0,1,1,0,0,1,0,0, 1,1,1,1,1,1,1,1, 0,0,0,0,0,1,1,1, 0,0,1,1,1,0,0,0, 1,1,0,0,1,0,0,0, 0,0,0,0,1,0,0,0,
 1,1,0,0,1,0,0,0, 0,0,0,0,1,0,0,0, 0,0,0,0,1,0,0,1, 1,0,0,1,0,0,1,0, 0,1,1,0,0,1,0,0, 1,1,1,1,1,1,1,1, 0,0,0,0,0,1,1,1, 0,0,1,1,1,0,0,0, 0,0,0,0,0,1,0,0,
 0,0,1,1,1,0,0,0, 1,1,0,0,1,0,0,0, 0,0,0,0,1,0,0,0, 0,0,0,0,1,0,0,1, 1,0,0,1,0,0,1,0, 0,1,1,0,0,1,0,0, 1,1,1,1,1,1,1,1, 0,0,0,0,0,1,1,1, 0,0,0,0,0,0,1,0,
 0,0,0,0,0,1,1,1, 0,0,1,1,1,0,0,0, 1,1,0,0,1,0,0,0, 0,0,0,0,1,0,0,0, 0,0,0,0,1,0,0,1, 1,0,0,1,0,0,1,0, 0,1,1,0,0,1,0,0, 1,1,1,1,1,1,1,1, 0,0,0,0,0,0,0,1,
};
*/

// The check matrix is used in the following equation :
//  error_syndrome = received_code * ecc_matrix_check
//     received_code is a vector of 72-elements [ 0 1 2 3 ... 70 71 ],
//       where bits 0..63 = data, and bits 64..71 = parity
//
//       bits are received LSB first,
//        0 1 2 3 4 5 6 7, 0 1 2 3 4 5 6 7, 0 1 2 3 4 5 6 7
static const unsigned char ecc_matrix_check[8*72] = 
{
1,0,1,0,0,1,0,0,
1,1,0,0,0,1,0,0,
1,1,0,0,0,0,1,0,
1,0,1,0,0,0,1,0,
1,0,0,1,1,1,1,0,
1,1,0,0,0,0,0,1,
1,0,1,0,0,0,0,1,
1,0,0,1,0,0,0,1,

0,1,0,1,0,0,1,0,
0,1,1,0,0,0,1,0,
0,1,1,0,0,0,0,1,
0,1,0,1,0,0,0,1,
0,1,0,0,1,1,1,1,
1,1,1,0,0,0,0,0,
1,1,0,1,0,0,0,0,
1,1,0,0,1,0,0,0,

0,0,1,0,1,0,0,1,
0,0,1,1,0,0,0,1,
1,0,1,1,0,0,0,0,
1,0,1,0,1,0,0,0,
1,0,1,0,0,1,1,1,
0,1,1,1,0,0,0,0,
0,1,1,0,1,0,0,0,
0,1,1,0,0,1,0,0,

1,0,0,1,0,1,0,0,
1,0,0,1,1,0,0,0,
0,1,0,1,1,0,0,0,
0,1,0,1,0,1,0,0,
1,1,0,1,0,0,1,1,
0,0,1,1,1,0,0,0,
0,0,1,1,0,1,0,0,
0,0,1,1,0,0,1,0,

0,1,0,0,1,0,1,0,
0,1,0,0,1,1,0,0,
0,0,1,0,1,1,0,0,
0,0,1,0,1,0,1,0,
1,1,1,0,1,0,0,1,
0,0,0,1,1,1,0,0,
0,0,0,1,1,0,1,0,
0,0,0,1,1,0,0,1,

0,0,1,0,0,1,0,1,
0,0,1,0,0,1,1,0,
0,0,0,1,0,1,1,0,
0,0,0,1,0,1,0,1,
1,1,1,1,0,1,0,0,
0,0,0,0,1,1,1,0,
0,0,0,0,1,1,0,1,
1,0,0,0,1,1,0,0,

1,0,0,1,0,0,1,0,
0,0,0,1,0,0,1,1,
0,0,0,0,1,0,1,1,
1,0,0,0,1,0,1,0,
0,1,1,1,1,0,1,0,
0,0,0,0,0,1,1,1,
1,0,0,0,0,1,1,0,
0,1,0,0,0,1,1,0,

0,1,0,0,1,0,0,1,
1,0,0,0,1,0,0,1,
1,0,0,0,0,1,0,1,
0,1,0,0,0,1,0,1,
0,0,1,1,1,1,0,1,
1,0,0,0,0,0,1,1,
0,1,0,0,0,0,1,1,
0,0,1,0,0,0,1,1,

1,0,0,0,0,0,0,0,
0,1,0,0,0,0,0,0,
0,0,1,0,0,0,0,0,
0,0,0,1,0,0,0,0,
0,0,0,0,1,0,0,0,
0,0,0,0,0,1,0,0,
0,0,0,0,0,0,1,0,
0,0,0,0,0,0,0,1
};

// syndrome_map[ syn ] 
//  The output of the syndrome_map is a 'bit-location' (#0 - #71) where
//  the decoder detects an error.  The "map" makes these assumptions
//  about the transmitted codeword : #0 - #63 = data, #64 - #71 = parity
//
//  For example, if the syndrome-value syn = 2 (decimal);  
//      syndrome_map [ syn=2 ] --> #65   ; // bit #65 is bad
//      syndrome_map [ syn=0 ] --> #254  ; // no error!
//      syndrome_map [ syn=3 ] --> #255  ; // uncorrectable!
//
static unsigned char syndrome_map[256] = {
254, 64,  65, 255,  66, 255, 255,  13,  67, 255, 255,  14, 255,  18,  21, 255,
 68,255, 255,  15, 255,  19,  22, 255, 255,  25,  26, 255,  29, 255, 255, 255,
 69,255, 255,   1, 255,   0,  23, 255, 255,  24,  27, 255,  30, 255, 255,  44,
255, 47,  33, 255,  34, 255, 255, 255,  37, 255, 255, 255, 255, 255, 255, 255,
 70,255, 255,   2, 255,   3,   9, 255, 255,  48,   8, 255,  31, 255, 255, 255,
255, 51,  32, 255,  35, 255, 255, 255,  38, 255, 255, 255, 255, 255,  52, 255,
255, 54,  55, 255,  41, 255, 255, 255,  42, 255, 255, 255, 255, 255, 255, 255,
 45,255, 255, 255, 255, 255, 255, 255, 255,   4, 255, 255, 255, 255, 255, 255,
 71,255, 255,   5, 255,   6,  10, 255, 255,   7,  11, 255,  17, 255, 255, 255,
255, 57,  56, 255,  16, 255, 255,  36,  39, 255, 255, 255, 255, 255, 255, 255,
255, 58,  59, 255,  40, 255, 255, 255,  43, 255, 255, 255, 255, 255, 255, 255,
 46,255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,  60, 255, 255, 255,
255, 61,  62, 255,  63, 255, 255, 255,  49, 255, 255,  28, 255, 255, 255, 255,
 50,255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
 53,255, 255, 255, 255,  20, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
255,255,  12, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255
};


// x = index for message-bits x = 0..(K-1)
// y = index for parity-bits x = 0..(N-K-1)
// z = index for 3rd-dimension of message-bits (depth) 0..(LEN-1)


// Generates parity-information from the *pMessage 
//     (ONLY the parity, not the entire codeword!)
// pMesage : a single 'frame' (1-plane) of userdata (one message bit 'x')
// pParityPlanes is an array of pointers.  pParityPlanes[ 0 ] = points
// to parity plane #0.

int 
ecc_generate_z1( eccword *pParityPlanes[], const eccword *pMessage, 
                const int x )
{
  int y, z;
  eccword *pParity;

  // validate parameters
  if ( pParityPlanes == NULL || pMessage == NULL )
  {
    ERROR_MESSAGE( "ecc_generate_z1() : pParityPlanes or pMessage == NULL!\n" )
    return ECCRC_FAIL;
  }

  for ( y = 0; y < ECC_NMK; ++y ) // # parity-bits (N-K)
  {
    pParity = pParityPlanes[ y ]; // get pointer to ParityPlane[y]

    // Generate an entire 'plane' of ECC-parity data.  
    if ( ecc_matrix_check[ y + (x*ECC_NMK) ] )
      for ( z = 0; z < ECC_Z; ++z )
        pParity[ z ] = pParity[ z ] ^ pMessage[ z ];
  }

  return ECCRC_OK;
}


// write 0's to all elements of all parityplanes.
// this function is called before the start of a new ECCblock
//   (encoder or decoder)
int 
ecc_clear_parityplanes( eccword *pParityPlanes[] )
{
  int y;

  // validate parameters
  if ( pParityPlanes == NULL )
  {
    ERROR_MESSAGE( "ecc_clear_parityplanes() : pParityPlanes == NULL!\n" )
    return ECCRC_FAIL;
  }

  for ( y = 0; y < ECC_NMK; ++y ) // # parity-bits (N-K)
  {
    memset( (void *) pParityPlanes[ y ], 0, ECC_Z * sizeof(eccword) );
  }

  return ECCRC_OK;
}


// calculate the error 'syndrome' from 1-plane of received userdata
// syndrome bits are organized horizontally
// syndrome = pSyndrome7[0], pSyndrome6[0], ... pSyndrome0[0]
//       (bit0 of pSyndrome7..pSyndrome0)
// pSydnromePlanes is an array of 8 planes.  pReceived is one-'x' bit of
//    the received codeword.
int 
ecc_decode_z1( eccword *pSyndromePlanes[], const eccword *pReceived,
              const int x )
{
  int y, z;
  eccword *pSyndrome;

  // validate parameters
  if ( pReceived == NULL || pSyndromePlanes == NULL )
  {
    ERROR_MESSAGE( "ecc_decode_z1() : pReceived or pSyndromePlanes == NULL!\n" )
    return ECCRC_FAIL;
  }

  for ( y = 0; y < ECC_NMK; ++y ) // # parity-bits (N-K)
  {
    pSyndrome = pSyndromePlanes[ y ];
    // generate ecc-bit #y, this is stored at pCodeword[ K+y ]
    if ( ecc_matrix_check[ y + (x*ECC_NMK) ] )
      for ( z = 0; z < ECC_Z; ++z )
      {
        pSyndrome[ z ] = pSyndrome[ z ] ^ pReceived[ z ];
      }
  }

  return ECCRC_OK;
}


// perform data correction operation
// pLocations is an array of 32 numbers; each corresponds to an error loc.
// pRecieved is an array of received planes

// since the codeword's parity-bits aren't needed by the application, we
// could change this equation to read "x < ECC_K"
//#define ECC_DO_CORRECTION( x ) ( (x) < ECC_N )
#define ECC_DO_CORRECTION( x ) ( (x) < ECC_K )

int 
ecc_correct_data32( eccword *pReceivedPlanes[], const unsigned char 
                   *pLocations, const int z )
{
  int i;

  eccword *pReceived;
  unsigned char location;
  eccword errorMask;

  // validate parameters
  if ( pReceivedPlanes == NULL || pLocations == NULL )
  {
    ERROR_MESSAGE( "ecc_correct_data32() : pReceivedPlanes or pLocations == NULL!\n" )
    return ECCRC_FAIL;
  }

  // i = bit-position within an eccword
//  for ( i = 0; i < 32; ++i ) // i < 32
  for ( i = 0; i < ECCWORD_BITWIDTH; ++i ) // i < 32
  {
    // generate a location by the precomputed table
    location = syndrome_map[ pLocations[ i ] ];
    if ( ECC_DO_CORRECTION( location ) )
    {
      // set pointer to corresponding bit 'x' (the error location)
      pReceived = pReceivedPlanes[ location ];

      // convert i (bit-position) into a 32-bit error-mask
      errorMask = 1 << i; // 000..010..000
      // perform a READ/MODIFY/WRITE
      pReceived[ z ] = pReceived[ z ] ^ errorMask;
    }

  }

  return ECCRC_OK;
}
 

// converts the 'syndrome pattern' into bit-locations
// syndrome bits are organized horizontally
// syndrome = pSyndrome7[0], pSyndrome6[0], ... pSyndrome0[0]
//       (bit0 of pSyndrome7..pSyndrome0)
// pReceivedPlanes, ECC_N planes of received codeword bits

//                   bit#
//     variable    0 1 2 3 ... 30 31
//     ---------------------------
//     pSyndrome0  a i q y                location0 = {hgfedcba}
//     pSyndrome1  b j r z                location1 = {ponmlkji}
//     pSyndrome2  c k s .                location2 = {xwvutsrq}
//     pSyndrome3  d l t .     syn_to_loc location3 = { ...  zy}
//     pSyndrome4  e m u .         ---->  location4 = ...
//     pSyndrome5  f n v                     ...
//     pSyndrome6  g o w                  location30=
//     pSyndrome7  h p x                  location31=
//

int 
ecc_syn_to_loc_z1( eccword *pReceivedPlanes[], eccword* const 
                  pSyndromePlanes[] )
{
  int x, y, z;
  eccword syndrome; // temp vars
  eccword accError; // error accumulator (used for short-cut evaluation)
  const eccword *pSyndrome;
  unsigned char locations[ECCWORD_BITWIDTH]; // 32 

  // validate parameters
  if ( pReceivedPlanes == NULL || pSyndromePlanes == NULL )
  {
    ERROR_MESSAGE( "ecc_syn_to_loc_z1() : pReceivedPlanes or pSyndromePlanes == NULL!\n" )
    return ECCRC_FAIL;
  }

  for ( z = 0; z < ECC_Z; ++z )
  {
    accError = 0; // clear this accumulator. 

//  for ( y = 0; y < ECC_NMK; ++y ) // # parity-bits (N-K)
    for ( y = ECC_NMK-1; y >= 0; --y ) // # parity-bits (N-K)
    {
      pSyndrome = pSyndromePlanes[ y ];
      syndrome = pSyndrome[z];
      accError = accError | syndrome; // or-sum the syndrome
      for ( x = 0; x < ECCWORD_BITWIDTH; ++x ) // # bits per word
      {
        locations[ x ] = (locations[ x ] << 1);
        locations[ x ] = locations[ x ] | 
                                (syndrome & 0x01); // bit #0
        syndrome = syndrome >> 1;  // shift right 1 bit
      }
    }
    // just finished computation of ECCWORD_BITWIDTH error-locations,
    // A zero-value for anyError indicates NO-ERRORs.  This knowledge
    // can be used as a shortcut.
    if ( accError != 0 )
      ecc_correct_data32( pReceivedPlanes, locations, z ); 
    //pLocations = pLocations + ECCWORD_BITWIDTH;  // advance pointer
  }

  return ECCRC_OK;
}


int 
ecc_decoder_shutdown( ecc_dec_state *pECCDecoder )
{
  int x,y;

  // validate parameters
  if ( pECCDecoder == NULL )
  {
    ERROR_MESSAGE( "ecc_decoder_shutdown() : pECCDecoder == NULL!\n" )
    return ECCRC_FAIL;
  }

  // deallocate RAM
  for ( x = 0; x < ECC_N; ++x )
    if ( pECCDecoder->pReceivedPlanes[x] != NULL )
    {
      free( pECCDecoder->pReceivedPlanes[x] );
      pECCDecoder->pReceivedPlanes[x] = NULL;
    }

  // deallocate RAM
  for ( y = 0; y < ECC_NMK; ++y )
    if ( pECCDecoder->pSyndromePlanes[y] != NULL )
    {
      free( pECCDecoder->pSyndromePlanes[y] );
      pECCDecoder->pSyndromePlanes[y] = NULL;
    }

  return ECCRC_OK;
}


int 
ecc_encoder_shutdown( ecc_enc_state *pECCEncoder )
{
  int y;

  // validate parameters
  if ( pECCEncoder == NULL )
  {
    ERROR_MESSAGE( "ecc_encoder_shutdown() : pECCEncoder == NULL!\n" )
    return ECCRC_FAIL;
  }

  // deallocate RAM
  if ( pECCEncoder->pMessagePlane != NULL )
  {
    free( pECCEncoder->pMessagePlane );
    pECCEncoder->pMessagePlane = NULL;
  }

  // deallocate RAM
  for ( y = 0; y < ECC_NMK; ++y )
    if ( pECCEncoder->pParityPlanes[y] != NULL )
    {
      free( pECCEncoder->pParityPlanes[y] );
      pECCEncoder->pParityPlanes[y] = NULL;
    }

  return ECCRC_OK;
}


int 
ecc_decoder_init( ecc_dec_state *pECCDecoder )
{
  int x,y;
  int flag_failure = 0;

  // validate parameters
  if ( pECCDecoder == NULL )
  {
    ERROR_MESSAGE( "ecc_decoder_init() : pECCDecoder == NULL!\n" )
    return ECCRC_FAIL;
  }

  ecc_decoder_shutdown( pECCDecoder );  // deallocate RAM, if necessary

  // reallocate RAM
  for ( x = 0; x < ECC_N; ++x )
  {
    pECCDecoder->pReceivedPlanes[x] = (eccword *) calloc( ((ECC_Z >> 2)+1), 
      (sizeof(eccword) << 2) );
    if ( pECCDecoder->pReceivedPlanes[x] == NULL )
      flag_failure = 1;
    else 
      memset( (void *) pECCDecoder->pReceivedPlanes[x], 0, 
              sizeof(eccword)*ECC_Z );
  }

  // reallocate RAM
  for ( y = 0; y < ECC_NMK; ++y )
  {
    pECCDecoder->pSyndromePlanes[y] = (eccword *) calloc( ((ECC_Z >> 2)+1), 
      (sizeof(eccword) << 2) );
    if ( pECCDecoder->pSyndromePlanes[y] == NULL )
      flag_failure = 1;
    else
      memset( (void *) pECCDecoder->pSyndromePlanes[y], 0, 
              sizeof(eccword)*ECC_Z );
  }

  // initialize statistic counters
  pECCDecoder->cCodeReceived = 0;
  pECCDecoder->cDataOut = 0;
  pECCDecoder->cTotalErrors = 0;
  pECCDecoder->cTotalCorrected = 0;

  if ( flag_failure )
    return ECCRC_FAIL;

  // clear the parity planes
  ecc_clear_parityplanes( pECCDecoder->pSyndromePlanes );

  return ECCRC_OK;
}


int 
ecc_encoder_init( ecc_enc_state *pECCEncoder )
{
  int y;
  int flag_failure = 0;

  // validate parameters
  if ( pECCEncoder == NULL )
  {
    ERROR_MESSAGE( "ecc_encoder_init() : pECCEncoder == NULL!\n" )
    return ECCRC_FAIL;
  }

  ecc_encoder_shutdown( pECCEncoder );  // deallocate RAM, if necessary

  // reallocate RAM
  pECCEncoder->pMessagePlane = (eccword *) calloc( ((ECC_Z >> 2)+1), 
    (sizeof(eccword) << 2) );
  if ( pECCEncoder->pMessagePlane == NULL )
    flag_failure = 1;
  else
    memset( (void *) pECCEncoder->pMessagePlane, 0, sizeof(eccword)*ECC_Z );


  // reallocate RAM
  for ( y = 0; y < ECC_NMK; ++y )
  {
    pECCEncoder->pParityPlanes[y] = (eccword *) calloc( ((ECC_Z >> 2)+1), 
      (sizeof(eccword) << 2) );
    if ( pECCEncoder->pParityPlanes[y] == NULL )
      flag_failure = 1;
    else
      memset( (void *) pECCEncoder->pParityPlanes[y], 0, 
              sizeof(eccword)*ECC_Z );
  }

  // initialize statistic counters
  pECCEncoder->cTotalEncoded = 0;
  pECCEncoder->cDataEncoded = 0;
  pECCEncoder->cParityOut = 0;

  if ( flag_failure )
    return ECCRC_FAIL;

  // clear the parity planes
  ecc_clear_parityplanes( pECCEncoder->pParityPlanes );

  return ECCRC_OK;
}


// encode 1-plane of userdata - the userdata is crunched into the
// encoder's parity-buffer.
// input : pMessagePlane 
int 
ecc_encoder_enc1( ecc_enc_state *pECCEncoder, const eccword *pMessagePlane )
{
  int x;
  int rc; // function return code

  // validate parameters
  if ( pECCEncoder == NULL || pMessagePlane == NULL )
  {
    ERROR_MESSAGE( "ecc_encoder_enc1() : pECCEncoder or pMessagePlane == NULL!\n" )
    return ECCRC_FAIL;
  }

  // check whether an ECCblock has just been completed
  if ( pECCEncoder->cParityOut == ECC_NMK )
  {
    // reset counters
    pECCEncoder->cDataEncoded = 0;
    pECCEncoder->cParityOut = 0;
  }

  // check 'position' of the encoder-state
  x = pECCEncoder->cDataEncoded;
  if ( x == ECC_K )
  {
    ERROR_MESSAGE( "ecc_encoder_enc1() : Already reached 1 'full' ECCblock!\n" )
    return ECCRC_FAIL;
  }
  else if ( x > ECC_K )
  {
    ERROR_MESSAGE( "ecc_encoder_enc1() : FATAL pECCEncoder->cDataEncoded > ECC_K !\n" )
    return ECCRC_FAIL;
  }


  // if this is FIRST input-data for the ECCblock, clear the Parity-buffer
  if ( x == 0 )
    rc = ecc_clear_parityplanes( pECCEncoder->pParityPlanes );

  // Input data : pMessagePlane,  Input bit-position 'x' = x
  // Update the parity-buffers (pECCEncoder->pParityPlanes)
  rc = ecc_generate_z1( pECCEncoder->pParityPlanes, pMessagePlane, x );

  // update counter
  if ( ECC_SUCCESS( rc ) )
    ++pECCEncoder->cDataEncoded;

  return rc;
}


// retrieve 1-plane of generated-parity from the encoder's parity-buffer.
// Output : returns a pointer to a pParityPlane
int 
ecc_encoder_getp1( ecc_enc_state *pECCEncoder, eccword *pParityPlane[] )
{
  int y;

  // validate parameters
  if ( pECCEncoder == NULL || pParityPlane == NULL )
  {
    ERROR_MESSAGE( "ecc_encoder_getp1() : pECCEncoder or pParityPlane == NULL!\n" )
    return ECCRC_FAIL;
  }

  // check 'position' of the encoder-state
  if ( pECCEncoder->cDataEncoded != ECC_K )
  {
    ERROR_MESSAGE( "ecc_encoder_getp1() : Not yet 1 'full' ECCblock!\n" )
    return ECCRC_FAIL;
  }

  y = pECCEncoder->cParityOut;
  if ( y == ECC_NMK )
  {
    ERROR_MESSAGE( "ecc_encoder_getp1() : Parity-buffer empty!\n" )
    return ECCRC_FAIL;
  }
  else if ( y > ECC_NMK )
  {
    ERROR_MESSAGE( "ecc_encoder_getp1() : FATAL pECCEncoder->cParityOut > ECC_NMK!\n" )
    return ECCRC_FAIL;
  }

  // get pointer to the correct ParityPlane
  *pParityPlane = pECCEncoder->pParityPlanes[ y ];

  // update counter
  ++pECCEncoder->cParityOut;

  return ECCRC_OK;
}


// retrieve pointer to a receive-buffer.  The client application can copy
// received-codeword data directly into the buffer, or simply zero out the
// provided receive-buffer.  (If the latter, the decoder will generate an
// 'error XOR map.')
// Output : returns a pointer to a pReceivePlane
int 
ecc_decoder_get_rcv_buffer( ecc_dec_state *pECCDecoder, eccword *pReceivePlane[] )
{
  int x;
  int rc; // function return code

  // validate parameters
  if ( pECCDecoder == NULL || pReceivePlane == NULL )
  {
    ERROR_MESSAGE( "ecc_decoder_get_rcv_buffer() : pECCDecoder or pReceivePlane == NULL!\n" )
    return ECCRC_FAIL;
  }

  // check whether an ECCblock has just been completely received
  if ( pECCDecoder->cDataOut == ECC_K )
  {
    // reset counters, clear the syndromeArray
    pECCDecoder->cCodeReceived = 0;
    pECCDecoder->cDataOut = 0;
    rc = ecc_clear_parityplanes( pECCDecoder->pSyndromePlanes );
  }

  // check 'position' of the decoder-state
  x = pECCDecoder->cCodeReceived;
  if ( x == ECC_N )
  {
    ERROR_MESSAGE( "ecc_decoder_get_rcv_buffer() : already have full ECCblock!\n" )
    return ECCRC_FAIL;
  }
  else if ( x > ECC_N )
  {
    ERROR_MESSAGE( "ecc_decoder_get_rcv_buffer() : FATAL pECCDecoder->cCodeReceived > ECC_N!\n" )
    return ECCRC_FAIL;
  }
 
  // get pointer to the correct ReceivedPlanes
  *pReceivePlane = pECCDecoder->pReceivedPlanes[ x ];

  return ECCRC_OK;
}


// process 1-plane of received-codeword through the ECC-decoder.  The
// decoder updates its internal syndrome-buffers.
int 
ecc_decoder_syn1( ecc_dec_state *pECCDecoder, const eccword *pReceivedPlane )
{
  int x;
  int rc; // function return code

  // validate parameters
  if ( pECCDecoder == NULL || pReceivedPlane == NULL )
  {
    ERROR_MESSAGE( "ecc_decoder_syn1() : pECCDecoder or pReceivedPlane == NULL!\n" )
    return ECCRC_FAIL;
  }

  // check whether an ECCblock has just been completely received
  if ( pECCDecoder->cDataOut == ECC_K )
  {
    // reset counters, clear the syndromeArray
    pECCDecoder->cCodeReceived = 0;
    pECCDecoder->cDataOut = 0;
    rc = ecc_clear_parityplanes( pECCDecoder->pSyndromePlanes );
  }

  // check 'position' of the decoder-state
  x = pECCDecoder->cCodeReceived;
  if ( x == ECC_N )
  {
    ERROR_MESSAGE( "ecc_decoder_syn1() : already have full ECCblock!\n" )
    return ECCRC_FAIL;
  }
  else if ( x > ECC_N )
  {
    ERROR_MESSAGE( "ecc_decoder_syn1() : FATAL pECCDecoder->cCodeReceived > ECC_N!\n" )
    return ECCRC_FAIL;
  }

  // Input data : pReceivedPlane,  Input bit-position 'x' = x
  // Update the syndrome-buffers (pECCDecoder->pSyndromePlanes)
  rc = ecc_decode_z1( pECCDecoder->pSyndromePlanes, pReceivedPlane, x );

  // update counter
  if ( ECC_SUCCESS( rc ) )
    ++pECCDecoder->cCodeReceived;

  return rc;
}


// perform error-correction operation -
//    flagDoCorrect = 0,  skip the correction step (don't do ECC process)
//    flagDoCorrect = 1,  perform ECC-correction)
// the correction-operation first calls syn_to_loc(), to convert the
// error-syndrome planes into error-location values.  The decoder
// then performs bitwise-XOR's on the marked error-locations (in the
// receive-buffer.)  If user-application placed the received-codeword
// into the receive-buffer, then the decoder will generate a corrected
// codeword.  On the other hand, if the user-app init'd the entire
// receive-buffer to 0, then the decoder will generate an error-map.
int 
ecc_decoder_correct( ecc_dec_state *pECCDecoder, const int flagDoCorrect )
{
  int x;
  int rc; // function return code

  // validate parameters
  if ( pECCDecoder == NULL )
  {
    ERROR_MESSAGE( "ecc_decoder_correct() : pECCDecoder == NULL!\n" )
    return ECCRC_FAIL;
  }

  // check 'position' of the decoder-state
  x = pECCDecoder->cCodeReceived;
  if ( x > ECC_N )
  {
    ERROR_MESSAGE( "ecc_decoder_correct() : FATAL pECCDecoder->cCodeReceived > ECC_N!\n" )
    return ECCRC_FAIL;
  }
  else if ( x != ECC_N )
  {
    ERROR_MESSAGE( "ecc_decoder_correct() : don't have full ECCblock yet!\n" )
    return ECCRC_FAIL;
  }

  if ( flagDoCorrect )
    rc = ecc_syn_to_loc_z1( pECCDecoder->pReceivedPlanes, 
        pECCDecoder->pSyndromePlanes );
  else
    rc = ECCRC_OK;  

  // update counter
  if ( ECC_SUCCESS( rc ) )
    pECCDecoder->cDataOut = 0;

  return ECCRC_OK;
}


// get decoded output
int 
ecc_decoder_get_out_buffer( ecc_dec_state *pECCDecoder, eccword **pDataPlane )
{
  int x;

  // validate parameters
  if ( pECCDecoder == NULL || pDataPlane == NULL )
  {
    ERROR_MESSAGE( "ecc_decoder_get_out_buffer() : pECCDecoder or pDataPlane == NULL!\n" )
    return ECCRC_FAIL;
  }

  // check 'position' of the decoder-state
  x = pECCDecoder->cCodeReceived;
  if ( x < ECC_N )
  {
    ERROR_MESSAGE( "ecc_decoder_get_out_buffer() : don't have full ECCblock yet!\n" )
    return ECCRC_FAIL;
  }
//else if ( x > ECC_N )
//{
//  ERROR_MESSAGE( "ecc_decoder_get_out_buffer() : FATAL pECCDecoder->cCodeReceived > ECC_N!\n" )
//  return ECCRC_FAIL;
//}

  // check 'position' of the decoder-state
  x = pECCDecoder->cDataOut;
  if ( x == ECC_K )
  {
    ERROR_MESSAGE( "ecc_decoder_get_out_buffer() : no data left!\n" )
    return ECCRC_FAIL;
  }
  else if ( x > ECC_K )
  {
    ERROR_MESSAGE( "ecc_decoder_get_out_buffer() : FATAL pECCDecoder->cDataOut > ECC_K!\n" )
    return ECCRC_FAIL;
  }
 
  // get pointer to the correct ReceivedPlanes
  *pDataPlane = pECCDecoder->pReceivedPlanes[ x ];

  // update counter
  ++pECCDecoder->cDataOut;

  return ECCRC_OK;
}
  
