

#ifndef __ECC_H
#define __ECC_H 1

// ECC_Z : USER ADJUSTABLE 
#define ECC_Z  (102524>>2)  // depth of a message-bit (#eccwords)

// NOT USER ADJUSTABLE : DO NOT CHANGE THE FOLLOWING PARAMETERS!
#define ECC_K 64  // mesage (data payload) length (#bits), per codeword
#define ECC_N 72  // total codeword length (#bits), per codeword
#define ECC_NMK ((ECC_N) - (ECC_K))  // parity length (#bits), per codeword

// utility macros
#define ERROR_MESSAGE( x )   { printf( "%s", x ); }
#define ECC_SUCCESS( rc )  ( (rc) == ECCRC_OK )

// constant definition 
#define ECCRC_OK   0  // generic success code
#define ECCRC_FAIL 1  // generic fail code

typedef unsigned int eccword; // basic unit of data (32-bit unsigned int)
#define ECCWORD_BITWIDTH 32   // 32-bits of data per 'eccword'

/////////////////////////////////////////////////////////////////////////////
//
// data structures
//

  typedef struct {
    eccword *pMessagePlane;   // memory, pointer to single MESSAGE plane 
    eccword *pParityPlanes[ECC_NMK]; // memory, pointer to 8 parity planes 
  
    int cDataEncoded; // count, #planes encoded so far (0..K)
    int cParityOut; // count, #parity planes read-out so far (0..NMK)

  
    // statistics
    int cTotalEncoded; // total # complete codewords encoded
  } ecc_enc_state; // ECC-encoder state



  typedef struct {
    eccword *pReceivedPlanes[ECC_N]; // memory, pointer to array of planes
    eccword *pSyndromePlanes[ECC_NMK]; // memory, pointer to 8 parity planes 
  
    int cCodeReceived; // count, #code planes received so far (0..N)
    int cDataOut; // count, #(user) data planes read-out so far (0..K)
  
    // statistics
    int cTotalErrors; // total # detected bit parity errors 
    int cTotalCorrected; // total #bit errors corrected
  } ecc_dec_state; // ECC-decoder state
  

/////////////////////////////////////////////////////////////////////////////
//
// function prototypes
//

  // *internal* private routines
  int ecc_generate_z1( eccword *pParityPlanes[], const eccword *pMessage, 
                const int x );
  int ecc_clear_parityplanes( eccword *pParityPlanes[] );
  int ecc_decode_z1( eccword *pSyndromePlanes[], const eccword *pReceived,
              const int x );
  int ecc_correct_data32( eccword *pReceivedPlanes[], const unsigned char 
                   *pLocations, const int z );
  int ecc_syn_to_loc_z1( eccword *pReceivedPlanes[], eccword * const 
                  pSyndromePlanes[] );

  // *public* interface routines

  // shutdown() : deallocates encoder's memory buffers
  int ecc_encoder_shutdown( ecc_enc_state *pECCEncoder );

  // init() : sets up encoder's memory buffers, clears all counters
  int ecc_encoder_init( ecc_enc_state *pECCEncoder );

  // enc1() : encode 1-plane of userdata -> parity
  int ecc_encoder_enc1( ecc_enc_state *pECCEncoder, 
                     const eccword *pMessagePlane );

  // get_p1() : retrieve 1-plane of generated-parity from encoder
  int ecc_encoder_getp1( ecc_enc_state *pECCEncoder, eccword *pParityPlane[] );

  //int ecc_encoder_update_status();

  // shutdown() : deallocates decoder's memory buffers
  int ecc_decoder_shutdown( ecc_dec_state *pECCDecoder );

  // init() : sets up decoder's memory buffers, clears all counters
  int ecc_decoder_init( ecc_dec_state *pECCDecoder );

  // get_rcv_buffer() : returns pointer to next-available receive-buffer,
  //                    to store 1-plane of the received codeword
  int ecc_decoder_get_rcv_buffer( ecc_dec_state *pECCDecoder, 
                               eccword *pReceivePlane[] );

  // get_rcv_buffer() : returns pointer to next-available output-buffer,
  //                    holds 1-plane of the received/corrected userdata
  int ecc_decoder_get_out_buffer( ecc_dec_state *pECCDecoder, 
                               eccword **pDataPlane );

  // syn1() : decode 1-plane of received codeword, updates decoder's syndrome
  //          table
  int ecc_decoder_syn1( ecc_dec_state *pECCDecoder, 
                     const eccword *pReceivedPlane );

  // correct() : if flagDoCorrect != 0, perform correction operation on 
  //             decoder's receive-buffer
  int ecc_decoder_correct( ecc_dec_state *pECCDecoder, 
                      const int flagDoCorrect );

//int ecc_decoder_update_status();








#endif // #ifndef __ECC_H
